#include "cg_manager.h"

#include <string>
#include <iostream>

using namespace std;

/***************************************************************************/

CG_Program::CG_Program(CGcontext context,
		       const string& fileName,
		       const string& procedureName,
			   CGenum programType,
		       CGprofile profile):
	profile_(profile)
{
	check_errors("Entering_CG_Program");
	if (!cgGLIsProfileSupported(profile_))
	{
		cerr << "cg-program " << procedureName
		     << " at file " << fileName	
		     << ": Selected profile is not supported." << endl;
		exit(1);
	}
	check_errors("CG_Program 1");

	program_ = cgCreateProgramFromFile(context, programType, fileName.c_str(),
					   profile_, procedureName.c_str(), NULL);
	check_errors("CG_Program 2");
	if (!cgIsProgramCompiled(program_))
		cgCompileProgram(program_);
	check_errors("CG_Program 3");

	cgGLEnableProfile(profile_);
	check_errors("CG_Program 4");
	cgGLLoadProgram(program_);
	check_errors("CG_Program 5");
	cgGLDisableProfile(profile_);
	check_errors("Leaving_CG_Program");
}

CGparameter CG_Program::getParameter(const string& paramName)
{
	return cgGetNamedParameter(program_, paramName.c_str());
}

void CG_Program::setParameter(const string& paramName, float x)
{
	cgGLSetParameter1f(getParameter(paramName), x);
}

void CG_Program::setParameter(const string& paramName, 
							float x, float y)
{
	cgGLSetParameter2f(getParameter(paramName),
			   x, y);
}

void CG_Program::setParameter(const string& paramName, 
					float x, float y, float z)
{
	cgGLSetParameter3f(getParameter(paramName),
			   x, y, z);
}

void CG_Program::setParameter(const string& paramName, 
					float x, float y, float z, float w)
{
	cgGLSetParameter4f(getParameter(paramName),
						x, y, z, w);
}

void CG_Program::setParameter(const string& paramName, 
			      CGGLenum matrix, CGGLenum xform)
{
	cgGLSetStateMatrixParameter(getParameter(paramName),
				    matrix, xform);
}

/*void CG_Program::setParameter(const string& paramName,
			      const Render_Buffer& buffer)
{
	cgGLSetTextureParameter(getParameter(paramName),
				buffer.getTextureId());
}*/

void CG_Program::setTexture(const string& paramName,
						const Render_Buffer& buffer)
{
	CGparameter param = getParameter(paramName);
	cgGLSetTextureParameter(param, buffer.getTextureId());
	cgGLEnableTextureParameter(param);

	//jor_step_middle.setParameter("fg", fg_buffer_);
	//cgGLEnableTextureParameter(jor_step_middle.getParameter("fg"));
}

void CG_Program::setTexture(const std::string& paramName, 
							GLuint texture)
{
	CGparameter param = getParameter(paramName);
	cgGLSetTextureParameter(param, texture);
	cgGLEnableTextureParameter(param);
}

void CG_Program::unsetTexture(const string& paramName)
{
	cgGLDisableTextureParameter(getParameter(paramName));

	//jor_step_middle.setParameter("fg", fg_buffer_);
	//cgGLEnableTextureParameter(jor_step_middle.getParameter("fg"));
}

void CG_Program::enable()
{
	cgGLEnableProfile(profile_);
	cgGLBindProgram(program_);
}

void CG_Program::disable()
{
	cgGLDisableProfile(profile_);
}

CG_Manager* CG_Manager::instance_ = 0;
CG_Manager& CG_Manager::getInstance()
{
	if (!instance_)
		instance_ = new CG_Manager();
	return *instance_;
}

void CG_Manager::loadProgram(const string& fileName,
				const string& procedureName,
				const string& programName,
				CGprofile profile)
{
	CG_Manager &manager = CG_Manager::getInstance();

	if (manager.programs_.find(programName) != 
		manager.programs_.end()) {
		cerr << "Program " << programName <<
				" already exists." << endl;
		exit(1);
	}	

	manager.programs_.insert(make_pair(programName,
				CG_Program(manager.context_,
							fileName, procedureName, CG_SOURCE, profile)));
	cout << "Loaded " << procedureName << endl;
}

void CG_Manager::loadAsmProgram(const string& fileName,
				const string& procedureName,
				const string& programName,
				CGprofile profile)
{
	CG_Manager &manager = CG_Manager::getInstance();

	if (manager.programs_.find(programName) != 
		manager.programs_.end()) {
		cerr << "Program " << programName <<
				" already exists." << endl;
		exit(1);
	}	

	manager.programs_.insert(make_pair(programName,
				CG_Program(manager.context_,
							fileName, procedureName, CG_OBJECT, profile)));
	cout << "Loaded " << procedureName << endl;
}

CG_Program CG_Manager::getProgram(const string& programName)
{
	CG_Manager &manager = CG_Manager::getInstance();

	map<string, CG_Program>::iterator result = manager.programs_.find(programName);

	if (result == manager.programs_.end()) {
		cerr << "Program " << programName << " not found." << endl;
		exit(1);
	}
	return result->second;
}
