#include <glui.h>
#include <GL/glut.h>
#include <GL/glaux.h>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>

#include <cstdio>

using namespace std;

////////////////////////////////////////////////////////////////////////////
// Types

typedef vector<vector<unsigned char> > charM;
enum { FLUID, WALL, CORRECTION, BOUNDARY };
enum { FILELOAD, FILESAVE };
enum { INFLOW, OUTFLOW, FREESLIP };
enum { NORTH, SOUTH, EAST, WEST };

struct BCondition {
	int ctype;
	int cdir;
	int start, finish;
	float u, v;
};

////////////////////////////////////////////////////////////////////////////
// Globals

float domainColors[4][3] = {
	{0.0, 146/256.0, 151/256.0},
	{1.0, 1.0, 1.0},
	{0.5, 0.0, 0.0},
	{1.0, 0.5, 0.0},
};

GLUI *glui;
int window;
charM domain;

GLUI_EditText *fileLoadEditText, *fileSaveEditText;
GLUI_StaticText *domainWidthStaticText, *domainHeightStaticText;
GLUI_Listbox *bcondTypes, *bcondDirs;
GLUI_EditText *xVel, *yVel;

string fileLoadName, fileSaveName;

vector<BCondition> boundaryConditions;

struct {
	float reynolds;
	float xLength;
	float yLength;
	float xVelocity;
	float yVelocity;
	float gridScale;
	int   startPos;
	int   finishPos;
	int   xSize;
	int   ySize;
	int   steps;
	float tau;
	int   lerp;
} uiParams;

////////////////////////////////////////////////////////////////////////////
// Loading the .bmp

bool fileExists(const char *filename)
{
	ifstream in(filename);
	return in.is_open();
}

void uncorrectImage()
{
	for (int i=0; i<domain.size(); i++)
		for (int j=0; j<domain[i].size(); j++)
			if (domain[i][j] == CORRECTION)
				domain[i][j] = FLUID;
}

void correctImage()
{
	bool valid = false;
	for (int i=0; i<domain.size(); i++) {
		domain[i][0] = WALL;
		domain[i][domain[0].size() - 1] = WALL;
	}
	for (int i=0; i<domain[0].size(); i++) {
		domain[0][i] = WALL;
		domain[domain.size() - 1][i] = WALL;
	}

	while (!valid) {
		valid = true;
		for (int i=1; i<domain.size()-1; i++)
			for (int j=1; j<domain[0].size()-1; j++)
				switch (domain[i][j]) {
					case WALL:
					case CORRECTION:
						if (domain[i][j-1] == FLUID &&
							domain[i][j+1] == FLUID) {
							domain[i][j-1] = CORRECTION;
							domain[i][j+1] = CORRECTION;
							valid = false;
						}
						if (domain[i-1][j] == FLUID &&
							domain[i+1][j] == FLUID) {
							domain[i-1][j] = CORRECTION;
							domain[i+1][j] = CORRECTION;
							valid = false;
						}
						break;
					case FLUID:
						if ((domain[i][j-1] != FLUID && domain[i][j+1] != FLUID) ||
							(domain[i-1][j] != FLUID && domain[i+1][j] != FLUID)) 
						{
							domain[i][j] = CORRECTION;
							valid = false;
						}
						break;
				}
	}
}

charM loadImage(const char *filename)
{
	AUX_RGBImageRec *rec = auxDIBImageLoad(filename);
	charM image = charM(rec->sizeY, vector<unsigned char>(rec->sizeX, ' '));

	for (int i=0; i<rec->sizeY; i++)
		for (int j=0; j<rec->sizeX; j++) {
			unsigned char r = rec->data[(i*rec->sizeX + j) * 3];
			unsigned char g = rec->data[(i*rec->sizeX + j) * 3 + 1];
			unsigned char b = rec->data[(i*rec->sizeX + j) * 3 + 2];

			int v = (r + g + b) / 3;

			if (v < 128)
				image[i][j] = WALL;
			else
				image[i][j] = FLUID;
		}

	return image;
}

////////////////////////////////////////////////////////////////////////////
// Display

void displayNothing()
{
	glClearColor(0.2, 0.2, 0.2, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glutSwapBuffers();
}

void displayDomain()
{
	glViewport(0, 0, glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//float m[16];
	//glGetFloatv(GL_PROJECTION_MATRIX, m);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	int w = domain[0].size(), h = domain.size();
	gluOrtho2D(0, w, 0, h);

	//glGetFloatv(GL_PROJECTION_MATRIX, m);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glDisable(GL_DEPTH_TEST);

	glBegin(GL_QUADS);
	for (int i=0; i<domain.size(); i++)
		for (int j=0; j<domain[i].size(); j++) {
			glColor3fv(domainColors[domain[i][j]]);
			glVertex3f(j  ,i  ,0);
			glVertex3f(j+1,i  ,0);
			glVertex3f(j+1,i+1,0);
			glVertex3f(j  ,i+1,0);
		}

	for (int i=0; i<boundaryConditions.size(); i++) {
		BCondition bc = boundaryConditions[i];
		glColor3fv(domainColors[BOUNDARY]);
		switch (bc.cdir) {
		case NORTH:
			glVertex3f(bc.start, domain.size() - 1, 0);
			glVertex3f(bc.finish + 1, domain.size() - 1, 0);
			glVertex3f(bc.finish + 1, domain.size(), 0);
			glVertex3f(bc.start, domain.size(), 0);
			break;
		case SOUTH:
			glVertex3f(bc.start, 0, 0);
			glVertex3f(bc.finish + 1, 0, 0);
			glVertex3f(bc.finish + 1, 1, 0);
			glVertex3f(bc.start, 1, 0);
			break;
		case EAST:
			glVertex3f(domain[0].size()-1, bc.start, 0);
			glVertex3f(domain[0].size(), bc.start, 0);
			glVertex3f(domain[0].size(), bc.finish + 1, 0);
			glVertex3f(domain[0].size()-1, bc.finish + 1, 0);
			break;
		case WEST:
			glVertex3f(0, bc.start, 0);
			glVertex3f(1, bc.start, 0);
			glVertex3f(1, bc.finish + 1, 0);
			glVertex3f(0, bc.finish + 1, 0);
			break;
		}
	}

	glEnd();

	glutSwapBuffers();
}

////////////////////////////////////////////////////////////////////////////
// GLUT Callbacks

void display()
{
	if (domain.size() == 0)
		displayNothing();
	else
		displayDomain();
}

void key(unsigned char k, int x, int y)
{
	if (k == 27)
		exit(0);
}

void setDomain(int x, int y, int v)
{
	int w = glutGet(GLUT_WINDOW_WIDTH),
		h = glutGet(GLUT_WINDOW_HEIGHT);
	y = h - y;

	float wr = w / (float) domain[0].size(),
		hr = h / (float) domain.size();

	float nx = x / wr, ny = y / wr;

	domain[(int)ny][(int)nx] = v;
}

void motionWall(int x, int y)
{
	setDomain(x, y, WALL);
	glutPostWindowRedisplay(window);
}

void motionFluid(int x, int y)
{
	setDomain(x, y, FLUID);
	glutPostWindowRedisplay(window);
}

void mouse(int button, int state, int x, int y)
{
	if (domain.size() == 0)
		return;

	switch (state) {
		case GLUT_UP:
			glutMotionFunc(NULL);
			break;
		case GLUT_DOWN:
			switch (button) {
				case GLUT_LEFT_BUTTON:
					glutMotionFunc(motionWall);
					motionWall(x, y);
					break;
				case GLUT_RIGHT_BUTTON:
					glutMotionFunc(motionFluid);
					motionFluid(x, y);
					break;
			}
	}
}

////////////////////////////////////////////////////////////////////////////
// GLUI Callbacks

void stringChange(int id)
{
	switch (id) {
		case FILELOAD:
			fileLoadName = fileLoadEditText->get_text();
			break;
		case FILESAVE:
			fileSaveName = fileSaveEditText->get_text();
			break;
	}
}

void loadBMPFromFile(int)
{
	if (fileExists(fileLoadName.c_str())) {
		domain = loadImage(fileLoadName.c_str());
		correctImage();
		uiParams.xSize = domain[0].size();
		uiParams.ySize = domain.size();
		glutSetWindow(window);
		glutReshapeWindow(768, 768 / (uiParams.xSize / (float) uiParams.ySize));		glutPostWindowRedisplay(window);
		glui->sync_live();
	}
}

void saveFLOToFile(int)
{
	ofstream out(fileSaveName.c_str());
	correctImage();

	out << "width " << uiParams.xSize-2 << endl;
	out << "height " << uiParams.ySize-2 << endl;
	out << "xlength " << uiParams.xLength << endl;
	out << "ylength " << uiParams.yLength << endl;
	out << "reynolds " << uiParams.reynolds << endl;
	out << "inksupersampling 1" << endl;
	out << "obstacle" << endl;
	for (int i=domain.size()-1; i>=0; i--) {
		for (int j=0; j<domain[i].size(); j++)
			out << (domain[i][j] == FLUID ? '.' : '*');
		out << endl;
	}
	for (int i=0; i<boundaryConditions.size(); i++) {
		BCondition bc = boundaryConditions[i];
		out << "boundary ";
		switch (bc.ctype) {
			case INFLOW:  out << "inflow "; break;
			case OUTFLOW: out << "outflow "; break;
			case FREESLIP: out << "freeslip "; break;
		}
		switch (bc.cdir) {
			case NORTH:  out << "north "; break;
			case SOUTH:  out << "south "; break;
			case WEST:  out << "west "; break;
			case EAST:  out << "east "; break;
		}
		out << bc.start << " " << bc.finish;
		if (bc.ctype == INFLOW)
			out << " " << bc.u << " " << bc.v;
		out << endl;
	}
	out << "gridscale " << uiParams.gridScale << endl;
	out << "steps " << uiParams.steps << endl;
	out << "tau " << uiParams.tau << endl;
	out << "lerp " << (uiParams.lerp ? "true" : "false") << endl;
}

void changedBCondType(int)
{
	switch (bcondTypes->get_int_val()) {
		case INFLOW:
			xVel->enable();
			yVel->enable();
			break;
		case OUTFLOW:
		case FREESLIP:
			xVel->disable();
			yVel->disable();
			break;
	}
}

void addBoundaryCondition(int)
{
	BCondition bc;
	bc.ctype = bcondTypes->get_int_val();
	bc.cdir = bcondDirs->get_int_val();
	bc.start = uiParams.startPos;
	bc.finish = uiParams.finishPos;
	bc.u = uiParams.xVelocity;
	bc.v = uiParams.yVelocity;

	boundaryConditions.push_back(bc);
	glutPostWindowRedisplay(window);
}

void clearBoundaryConditions(int)
{
	boundaryConditions.clear();
	glutPostWindowRedisplay(window);
}

void correctDomain(int)
{
	correctImage();
	glutPostWindowRedisplay(window);
}

void uncorrectDomain(int)
{
	uncorrectImage();
	glutPostWindowRedisplay(window);
}

void cleanDomain(int)
{
	domain = charM(uiParams.ySize, vector<unsigned char>(uiParams.xSize, FLUID));
	cout << domain.size() << " " << domain[0].size() << endl;
	correctImage();
	clearBoundaryConditions(0);

	glutSetWindow(window);
	glutReshapeWindow(768, 768 / (uiParams.xSize / (float) uiParams.ySize));
	if (uiParams.xSize > uiParams.ySize) {
		uiParams.yLength = 1.0;
		uiParams.xLength = (uiParams.xSize / (float) uiParams.ySize);
	} else {
		uiParams.xLength = 1.0;
		uiParams.yLength = (uiParams.ySize / (float) uiParams.xSize);
	}
	glui->sync_live();
	glutPostWindowRedisplay(window);
}

////////////////////////////////////////////////////////////////////////////
// Initialization

void initGLUI()
{
	glui = GLUI_Master.create_glui( "GLUI" );
	GLUI_Master.set_glutIdleFunc(NULL);
	GLUI_Panel *flowParams = glui->add_panel("Flow parameters");

	glui->add_edittext_to_panel(flowParams, "Reynolds number:",	
					GLUI_EDITTEXT_FLOAT, &uiParams.reynolds)
		->set_float_limits(0, 100000);
	glui->add_edittext_to_panel(flowParams, "x Length:",			GLUI_EDITTEXT_FLOAT, &uiParams.xLength)
		->set_float_limits(0, 100000);
	glui->add_edittext_to_panel(flowParams, "y Length:",			GLUI_EDITTEXT_FLOAT, &uiParams.yLength)
		->set_float_limits(0, 100000);
	glui->add_edittext_to_panel(flowParams, "Grid Scale:",		GLUI_EDITTEXT_FLOAT, &uiParams.gridScale)
		->set_float_limits(0, 100000);

	GLUI_Panel *loadBMP = glui->add_panel("Loading images");
	fileLoadEditText = glui->add_edittext_to_panel(loadBMP, "File Name:",
		GLUI_EDITTEXT_TEXT, NULL, FILELOAD, stringChange);
	glui->add_button_to_panel(loadBMP, "Load", 0, loadBMPFromFile);

	GLUI_Panel *saveFLO = glui->add_panel("Saving problem files");
	fileSaveEditText = glui->add_edittext_to_panel(saveFLO, "File Name:",
		GLUI_EDITTEXT_TEXT, NULL, FILESAVE, stringChange);
	glui->add_button_to_panel(saveFLO, "Save", 0, saveFLOToFile);

	GLUI_Panel *domainProps = 
		glui->add_panel("Domain properties");
	glui->add_edittext_to_panel(domainProps, "Width:", GLUI_EDITTEXT_INT, &uiParams.xSize);
	glui->add_edittext_to_panel(domainProps, "Height:", GLUI_EDITTEXT_INT, &uiParams.ySize);
	glui->add_edittext_to_panel(domainProps, "Tau:", GLUI_EDITTEXT_FLOAT, &uiParams.tau)
		->set_float_limits(0.01f, 1.0f);
	glui->add_edittext_to_panel(domainProps, "Steps per time adjust:", GLUI_EDITTEXT_INT, &uiParams.steps)
		->set_int_limits(1, 1000000);
	glui->add_checkbox_to_panel(domainProps, "Lerp ink:", &uiParams.lerp);
	uiParams.steps = 2;
	uiParams.tau = 0.5;
	uiParams.lerp = false;
	glui->add_button_to_panel(domainProps, "Create new domain", 0, cleanDomain);

	GLUI_Panel *boundaryConds =
		glui->add_panel_to_panel(domainProps, "Boundary conditions");
	
	bcondTypes = glui->add_listbox_to_panel(boundaryConds, "Type:", NULL, -1, changedBCondType);
	bcondTypes->add_item(INFLOW, "Inflow");
	bcondTypes->add_item(OUTFLOW, "Outflow");
	bcondTypes->add_item(FREESLIP, "Free-Slip");

	bcondDirs = glui->add_listbox_to_panel(boundaryConds, "Direction:");
	bcondDirs->add_item(NORTH, "North");
	bcondDirs->add_item(SOUTH, "South");
	bcondDirs->add_item(EAST, "East");
	bcondDirs->add_item(WEST, "West");

	xVel = glui->add_edittext_to_panel(boundaryConds, "x Velocity", GLUI_EDITTEXT_FLOAT,
		&uiParams.xVelocity);
	yVel = glui->add_edittext_to_panel(boundaryConds, "y Velocity", GLUI_EDITTEXT_FLOAT,
		&uiParams.yVelocity);

	glui->add_edittext_to_panel(boundaryConds, "start", GLUI_EDITTEXT_INT, &uiParams.startPos);
	glui->add_edittext_to_panel(boundaryConds, "end", GLUI_EDITTEXT_INT, &uiParams.finishPos);

	glui->add_button_to_panel(boundaryConds, "Add", -1, addBoundaryCondition);
	glui->add_button_to_panel(boundaryConds, "Clear", -1, clearBoundaryConditions);

	glui->add_button("Correct Domain", 0, correctDomain);
	glui->add_button("Erase corrections", 0, uncorrectDomain);
	glui->sync_live();

	
}

void initGLUT()
{
	int argc = 1;
	char *argv[1] = {"GLUT"};
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowPosition(0,0);
	glutInitWindowSize(512,512);

	window = glutCreateWindow("floloader");
	glEnable(GL_SMOOTH);
	glDisable(GL_LIGHTING);
	glViewport(0, 0, 512, 512);

	glutDisplayFunc(display);
	glutKeyboardFunc(key);
	glutMouseFunc(mouse);
}

void init()
{
	initGLUT();
	initGLUI();
}

////////////////////////////////////////////////////////////////////////////

void setup()
{
	uiParams.gridScale = 8.0;
	uiParams.xLength = 1.0;
	uiParams.yLength = 1.0;
	uiParams.reynolds = 100.0;
	uiParams.xSize = 64;
	uiParams.ySize = 64;
	glui->sync_live();
}

////////////////////////////////////////////////////////////////////////////

void run()
{
	glutMainLoop();
}

////////////////////////////////////////////////////////////////////////////

void main(int argc, char **argv)
{
	init();
	setup();
	run();
}