#include "navier_stokes.h"
#include "cg_manager.h"
#include "glgpu.h"

#include <Cg/cgGL.h>
#include <cassert>
#include <cmath>

#include <iostream>
#include <fstream>
#include <string>

using namespace std;

#define TURBO_MODE_ENABLED

/*****************************************************************************/
// Dumb parsing stuff

static InkQuality inkQualityFromString(const string& str)
{
	if ("float" == str) return iqFloating;
	if ("fixed" == str) return iqFixed;
	cerr << "Parse error: '" << str << "' not a valid ink quality setting" << endl;
	exit(1);
}

static Direction directionFromString(const string& str)
{
	if (str == "north")	return north;
	else if (str == "south") return south;
	else if (str == "east") return east;
	else if (str == "west") return west;
	else {
		cerr << "parse error: bad direction" << endl;
		exit(1);
	}
}

static int colorSizeFromString(const string& str)
{
	if (str == "red") return SPLAT_RED;
	else if (str == "green") return SPLAT_GREEN;
	else if (str == "blue") return SPLAT_BLUE;
	else if (str == "cyan") return SPLAT_CYAN;
	else if (str == "magenta") return SPLAT_MAGENTA;
	else if (str == "yellow") return SPLAT_YELLOW;
	else if (str == "black") return SPLAT_BLACK;
	else if (str == "white") return SPLAT_WHITE;
	else if (str == "small") return SPLAT_SMALL;
	else if (str == "medium") return SPLAT_MEDIUM;
	else if (str == "large") return SPLAT_LARGE;
	else if (str == "huge") return SPLAT_HUGE;
	else {
		cerr << "Parse error: bad color or size\n";
		exit(1);
	}
}

/*****************************************************************************/
// Splats, etc

void Navier_Stokes_Solver::splat_choose(int choice)
{
	float colors[SPLAT_COLORS][4] = {
		{1.0f, 0.0f, 0.0f, 1.0f},
		{0.0f, 1.0f, 0.0f, 1.0f},
		{0.0f, 0.0f, 1.0f, 1.0f},
		{0.0f, 1.0f, 1.0f, 1.0f},
		{1.0f, 0.0f, 1.0f, 1.0f},
		{1.0f, 1.0f, 0.0f, 1.0f},
		{0.0f, 0.0f, 0.0f, 1.0f},
		{1.0f, 1.0f, 1.0f, 1.0f},
	};

	switch (choice) {
		case SPLAT_SMALL:
			splatRadius = 1.0f;
			break;
		case SPLAT_MEDIUM:
			splatRadius = 3.0f;
			break;
		case SPLAT_LARGE:
			splatRadius = 5.0f;
			break;
		case SPLAT_HUGE:
			splatRadius = Width() / 5.0f;
			break;
		default:
			assert(choice < SPLAT_COLORS);
			splatColor[0] = colors[choice][0];
			splatColor[1] = colors[choice][1];
			splatColor[2] = colors[choice][2];
			splatColor[3] = colors[choice][3];
			break;
	}
}

Splat Navier_Stokes_Solver::get_custom_splat()
{
	Splat s;
	s.color_[0] = splatColor[0];
	s.color_[1] = splatColor[1];
	s.color_[2] = splatColor[2];
	s.color_[3] = splatColor[3];
	s.radius_ = splatRadius;

	return s;
}

/*****************************************************************************/
// r_buffer_width and r_buffer_height compute the _r_educe buffer size
// a reduce buffer is larger than usual because it has to store all the 
// levels (inneficient but fast)

static int r_buffer_width(int height, int width)
{
	int m = max(height, width);

	int v = 1;
	while (v < m) v <<= 1;

	return v + (v >> 1);
}

static int r_buffer_height(int height, int width)
{
	int m = max(height, width);

	int v = 1;
	while (v < m) v <<= 1;

	return v;
}

void Navier_Stokes_Solver::init_ibfv()
{
	check_errors("before void Navier_Stokes_Solver::init_ibfv()");
	unsigned char **noise = new unsigned char*[noise_texture_count_];
	for (int j=0; j<noise_texture_count_; ++j) {
		noise[j] = new unsigned char[domainHeight_ * domainWidth_ * 
				inkSuperSampling_ * inkSuperSampling_];
	}
	for (int i=0; i<(domainHeight_*domainWidth_*
					inkSuperSampling_*inkSuperSampling_); i++)
	{
		for (int j=0; j<noise_texture_count_; ++j) {
			int r = (int)(((float) rand() / (float) RAND_MAX) * 256);
			if (r < 127)
				noise[j][i] = 0;
			else
				noise[j][i] = 255;
		}
	}

	for (int j=0; j<noise_texture_count_; ++j) {
		glEnable(GL_TEXTURE_RECTANGLE_NV);
		glBindTexture(GL_TEXTURE_RECTANGLE_NV, noise_texture_[j]);
		glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_LUMINANCE, domainWidth_ * inkSuperSampling_,
			domainHeight_ * inkSuperSampling_, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, noise[j]);
		delete[] noise[j];
	}
	delete[] noise;
	check_errors("after void Navier_Stokes_Solver::init_ibfv()");
}

void Navier_Stokes_Solver::init_buffers()
{
	check_errors("Before void Navier_Stokes_Solver::init_buffers()");
	fg_buffer_  = new Render_Buffer("float=32 rg textureRECT", domainWidth_, domainHeight_, false);
	// b stores poisson rhs

	uv_buffer_  = new Render_Buffer("double float=32 rgba textureRECT", domainWidth_, domainHeight_, false);
	if (iqFloating == inkQuality_) {
		cout << "Floating-point ink buffer selected" << endl;
		ink_buffer_ = new Render_Buffer("double float=16 rgba textureRECT",
						inkSuperSampling_*domainWidth_,
						inkSuperSampling_*domainHeight_, false);
	} else {
		cout << "Fixed-point ink buffer selected" << endl;
		ink_buffer_ = new Render_Buffer("double rgba textureRECT",
						inkSuperSampling_*domainWidth_,
						inkSuperSampling_*domainHeight_, false);
	}

	p_buffer_    = new Render_Buffer("float=32 r textureRECT", domainWidth_, domainHeight_, false);
	//r_buffer_    = new Render_Buffer("float=32 rgba textureRECT", r_buffer_width(domainWidth_, domainHeight_),
					//r_buffer_height(domainWidth_, domainHeight_), false);
	r_buffer_    = new Render_Buffer("double float=32 rgba textureRECT", domainWidth_, domainHeight_, false);
	//flag_buffer_ = new Render_Buffer("float=32 rgba textureRECT", domainWidth_, domainHeight_, false);
	obstacle_grid_ = vector<vector<unsigned char> >(2+domainHeight_, vector<unsigned char>(2+domainWidth_, OBSTACLE));

	glEnable(GL_TEXTURE_RECTANGLE_NV);
	glGenTextures(1, &flag_texture_);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, flag_texture_);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_FLOAT_RGBA32_NV, domainWidth_, domainHeight_, 0, GL_RGBA, GL_FLOAT, NULL);

	glGenTextures(1, &obstacle_texture_);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, obstacle_texture_);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 4, domainWidth_, domainHeight_, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

	glGenTextures(1, &neighbors_texture_);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, neighbors_texture_);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 4, domainWidth_, domainHeight_, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

	glGenTextures(noise_texture_count_, noise_texture_);
	for (int i=0; i<noise_texture_count_; ++i) {
		glBindTexture(GL_TEXTURE_RECTANGLE_NV, noise_texture_[i]);
		glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_LUMINANCE,
				domainWidth_ * inkSuperSampling_,
				domainHeight_ * inkSuperSampling_,
				0, GL_LUMINANCE, GL_UNSIGNED_BYTE, NULL);
	}

	// create default obstacle
	// obstacle is simply a string of asterisks and points, 
	// left-to-right, top-to-bottom, one char per pixel

	string border(domainWidth_, '*'), middle(domainWidth_ - 2, '.'), o;
	middle = "*" + middle + "*";
	o = border;
	for (int i=0; i<domainHeight_-2; i++)
		o = o + middle;
	o = o + border;
	set_obstacle_grid(o.c_str());
	set_obstacle();
	
	check_errors("After void Navier_Stokes_Solver::init_buffers()");
}

void Navier_Stokes_Solver::init_solver()
{
	init_buffers();
	init_ibfv();

	dx_ = xLength_/domainWidth_;
	dy_ = yLength_/domainHeight_;
	dt_ = 0.02f;
	tau_ = 0.8f;
	steps_per_time_adjustment_ = 1;
	max_u_ = 0.0f;
	max_v_ = 0.0f;
	totalTime_ = 0.0f;
	lerp_ = false;
	ibfv_ = false;
	heated_ = false;
	currentIbfvAlpha_ = 1.0f;
	advection_ = true;


	gravity_x_ = gravity_y_ = 0.0f;
	load_all_programs();
	setup_boundary_conditions();

	ink_to   = GL_FRONT;
	ink_from = GL_BACK;

	pressure_s1 = GL_FRONT;
	pressure_s2 = GL_FRONT;

	clear_ink();

	// We need fences to avoid flushing the pipeline in a reduction
	glGenFencesNV(1, &fence_);

	dump_ = false;
	total_frame_counter_ = 1;
	frame_counter_ = 0;
}

/*****************************************************************************/
// Construction

Navier_Stokes_Solver::Navier_Stokes_Solver(int domainWidth,
					   int domainHeight,
					   float xLength,
					   float yLength,
					   float reynoldsNumber,
					   int inkSuperSampling):
	domainWidth_(domainWidth + 2), 
	domainHeight_(domainHeight + 2),
	xLength_(xLength),
	yLength_(yLength),
	reynoldsNumber_(reynoldsNumber),
	inkSuperSampling_(inkSuperSampling)
{
	turboInk_ = 1;
	inkTime_ = pressureTime_ = timestepTime_ = fgTime_ = boundaryTime_ = newuvTime_ = 0;
	inkQuality_ = iqFloating;
	init_solver();
}

/*****************************************************************************/
// Obstacle Handling

// Some observations here:
// Every time there is an offset, the multiplier is -1. We use this in obstacle.cg
// to set the adiabatic conditions on the walls

float obstacle_values[13][4] = 
	{
		// The commented values are implicit in the fragment shader,
		// xofx  xofy  xm  yofx  yofy  ym  
		{ /*0,*/ 1,   -1,  0,  /*0,*/  0}, // EN   // must set temperature equal to north
		{ /*0,*/ 0,    0,  0,  /*0,*/  0}, // CNE  
		{ /*0,*/ 0,    0,  1,  /*0,*/ -1}, // EE   // must set temperature equal to east
		{ /*0,*/ 0,    0,  1,  /*0,*/ -1}, // CSE
		{ /*0,*/-1,   -1,  0,  /*0,*/  1}, // ES   // must set temperature equal to south
		{ /*0,*/ 0,    1,  0,  /*0,*/  1}, // CSW
		{ /*0,*/ 0,    1, -1,  /*0,*/ -1}, // EW   // must set temperature equal to west
		{ /*0,*/ 1,   -1,  0,  /*0,*/  0}, // CNW
		{ /*0,*/ 0,    1,  0,  /*0,*/  0}, // FS
		{ /*0,*/ 0,    0,  0,  /*0,*/  0}, // FSW
		{ /*0,*/ 0,    0,  0,  /*0,*/  1}, // FW
		{ /*0,*/ 0,    1,  0,  /*0,*/  1}, // FLUID
		{ /*0,*/ 0,    0,  0,  /*0,*/  0}  // WALL
	};

void Navier_Stokes_Solver::set_obstacle_grid(const char *grid)
{
	check_errors("before void Navier_Stokes_Solver::set_obstacle_grid(const char *grid)");
	for (int i=1; i<=domainHeight_; i++)
		for (int j=1; j<=domainWidth_; j++) {
			char x = *grid++;
			if (x == '*') {
                obstacle_grid_[i][j] = OBSTACLE;
			}
			else {
				obstacle_grid_[i][j] = NOT_OBSTACLE;
			}
		}
	check_errors("after void Navier_Stokes_Solver::set_obstacle_grid(const char *grid)");
}

void Navier_Stokes_Solver::set_obstacle()
{
	check_errors("before void Navier_Stokes_Solver::set_obstacle()");
	float *v = new float[domainHeight_ * domainWidth_ * 4];
	float *b = v;
	unsigned char *obs = new unsigned char[domainHeight_ * domainWidth_ * 4],
				  *neighbors = new unsigned char[domainHeight_ * domainWidth_ * 4];
	unsigned char *bobs = obs,
				  *bn = neighbors;
	
	for (int i=1; i<=domainHeight_; i++)
		for (int j=1; j<=domainWidth_; j++)
			if (obstacle_grid_[i][j] == OBSTACLE) {
				*bobs++ = 0;
				*bobs++ = 0;
				*bobs++ = 0;
				*bobs++ = 0;
			} else {
				*bobs++ = 255;
				*bobs++ = 255;
				*bobs++ = 255;
				*bobs++ = 255;
			}

	glEnable(GL_TEXTURE_RECTANGLE_NV);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, obstacle_texture_);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 4, domainWidth_, domainHeight_, 0, GL_RGBA, GL_UNSIGNED_BYTE, obs);
	
	for (int i=1; i<=domainHeight_; i++) {
		for (int j=1; j<=domainWidth_; j++) {
			OBSTACLE_FLAG fl;
			char me = obstacle_grid_[i][j],
				 n = obstacle_grid_[i+1][j],
				 s = obstacle_grid_[i-1][j],
				 w = obstacle_grid_[i][j-1],
				 e = obstacle_grid_[i][j+1];
			
			*bn++ = (n == OBSTACLE) ? 255 : 0; // neighbor r
			*bn++ = (s == OBSTACLE) ? 255 : 0; // neighbor g
			*bn++ = (e == OBSTACLE) ? 255 : 0; // neighbor b
			*bn++ = (w == OBSTACLE) ? 255 : 0; // neighbor a
				 
			if (me == NOT_OBSTACLE) { // FLUID
				if (n == OBSTACLE && e == OBSTACLE)
					fl = FSW;
				else if (n == OBSTACLE)
					fl = FS;
				else if (e == OBSTACLE)
					fl = FW;
				else
					fl = FLUID;
			} else { // WALL
				if      (n == NOT_OBSTACLE && e == OBSTACLE  && w == OBSTACLE  && s == OBSTACLE)
					fl = EN;
				else if (n == OBSTACLE  && e == NOT_OBSTACLE && w == OBSTACLE  && s == OBSTACLE)
					fl = EE;
				else if (n == OBSTACLE  && e == OBSTACLE  && w == NOT_OBSTACLE && s == OBSTACLE)
					fl = EW;
				else if (n == OBSTACLE  && e == OBSTACLE  && w == OBSTACLE  && s == NOT_OBSTACLE)
					fl = ES;
				else if (n == OBSTACLE  && e == OBSTACLE  && s == NOT_OBSTACLE && w == NOT_OBSTACLE)
					fl = CSW;
				else if (n == OBSTACLE  && e == NOT_OBSTACLE  && s == NOT_OBSTACLE && w == OBSTACLE)
					fl = CSE;
				else if (n == NOT_OBSTACLE && e == NOT_OBSTACLE && s == OBSTACLE  && w == OBSTACLE)
					fl = CNE;
				else if (n == NOT_OBSTACLE && e == OBSTACLE  && s == OBSTACLE  && w == NOT_OBSTACLE)
					fl = CNW;
				else if (n == OBSTACLE  && e == OBSTACLE  && s == OBSTACLE  && w == OBSTACLE)
					fl = WALL;
				else {
					cout << "Ill-formed obstacle!" << endl;
					exit(1);
				}
			}
			*b++ = obstacle_values[fl][0];
			*b++ = obstacle_values[fl][1];
			*b++ = obstacle_values[fl][2];
			*b++ = obstacle_values[fl][3];
		}
	}

	glEnable(GL_TEXTURE_RECTANGLE_NV);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, flag_texture_);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_FLOAT_RGBA32_NV, domainWidth_, domainHeight_, 0, GL_RGBA, GL_FLOAT, v);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, neighbors_texture_);
	glTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 4, domainWidth_, domainHeight_, 0, GL_RGBA, GL_UNSIGNED_BYTE, neighbors);
	delete[] v;
	delete[] obs;
	delete[] neighbors;
	
	check_errors("after void Navier_Stokes_Solver::set_obstacle()");
}

/*****************************************************************************/
// Boundary conditions

void Navier_Stokes_Solver::add_boundary_condition(Direction dir, Type type)
{
	ConditionDirection *cd;
	ConditionType *ct;
	int i;

	switch (dir) {
		case north:	cd = new NorthCondition(1, domainWidth_-1, domainHeight_); break;
		case south:	cd = new SouthCondition(1, domainWidth_-1); break;
		case east:	cd = new EastCondition(1, domainHeight_-1, domainWidth_);	break;
		case west:	cd = new WestCondition(1, domainHeight_-1);	break;
		default:
			exit(1);
	}

	switch (type) {
		case noSlip:   ct = new NoSlip();	break;
		case freeSlip: ct = new FreeSlip(); break;
		case outFlow: 
			ct = new OutFlow();  
			switch (cd->direction_) {
				case north:
					for (i=1; i<domainWidth_-1; i++)
						obstacle_grid_[domainHeight_][i] = NOT_OBSTACLE;
					break;
				case south:
					for (i=1; i<domainWidth_-1; i++)
						obstacle_grid_[1][i] = NOT_OBSTACLE;
					break;
				case east:
					for (i=1; i<domainHeight_-1; i++)
						obstacle_grid_[i][domainWidth_] = NOT_OBSTACLE;
					break;
				case west:
					for (i=1; i<domainHeight_-1; i++)
						obstacle_grid_[i][1] = NOT_OBSTACLE;
					break;
				default:
					cerr << "Invalid direction in void Navier_Stokes_Solver::add_boundary_condition(Direction dir, Type type)" << endl;
					exit(1);
			}
			break;
		default:
			exit(1);
	}

	conditions_.push_front(new UVCondition(cd, ct));
	set_obstacle();
}

/*****************************************************************************/

void Navier_Stokes_Solver::add_boundary_condition(Direction dir, Type type, int start, int finish)
{
	ConditionDirection *cd;
	ConditionType *ct;
	int i, mn, mx;
	switch (dir) {
		case north:	cd = new NorthCondition(start, finish, domainHeight_); break;
		case south:	cd = new SouthCondition(start, finish); break;
		case east:	cd = new EastCondition(start, finish, domainWidth_);	break;
		case west:	cd = new WestCondition(start, finish);	break;
		default:
			exit(1);
	}

	switch (type) {
		case noSlip:   ct = new NoSlip();	break;
		case freeSlip: ct = new FreeSlip(); break;
		case outFlow:  
			ct = new OutFlow();  
			mn = min(start, finish)+1;
			mx = max(start, finish)+1;
			switch (cd->direction_) {
				case north:
					for (i=mn; i<=mx; i++)
						obstacle_grid_[domainHeight_][i] = NOT_OBSTACLE;
					break;
				case south:
					for (i=mn; i<=mx; i++)
						obstacle_grid_[1][i] = NOT_OBSTACLE;
					break;
				case east:
					for (i=mn; i<=mx; i++)
						obstacle_grid_[i][domainWidth_] = NOT_OBSTACLE;
					break;
				case west:
					for (i=mn; i<=mx; i++)
						obstacle_grid_[i][1] = NOT_OBSTACLE;
					break;
				default:
					cerr << "Invalid direction in void Navier_Stokes_Solver::add_boundary_condition(Direction dir, Type type)" << endl;
					exit(1);
			}
			break;
		default:
			cerr << "Invalid enumerant in void Navier_Stokes_Solver::add_boundary_condition(Direction dir, Type type, int start, int finish)" << endl;
			exit(1);
	}

	conditions_.push_front(new UVCondition(cd, ct));
	set_obstacle();
}

/*****************************************************************************/

#define HEAT_TEST

void Navier_Stokes_Solver::add_boundary_condition(Direction dir, float u, float v, float t)
{
	ConditionDirection *cd;
	int i;

	switch (dir) {
		case north:	cd = new NorthCondition(1, domainWidth_-1, domainHeight_); 
#ifndef HEAT_TEST
					for (i=1; i<domainWidth_; i++)
						obstacle_grid_[domainHeight_][i] = NOT_OBSTACLE;
#endif
					break;
		case south:	cd = new SouthCondition(1, domainWidth_-1); break;
#ifndef HEAT_TEST
					for (i=1; i<domainWidth_; i++)
						obstacle_grid_[1][i] = NOT_OBSTACLE;
#endif
		case east:	cd = new EastCondition(1, domainHeight_-1, domainWidth_);	
#ifndef HEAT_TEST
					for (i=1; i<domainHeight_; i++)
						obstacle_grid_[i][domainWidth_] = NOT_OBSTACLE;
					break;
#endif
		case west:	cd = new WestCondition(1, domainHeight_-1);	
#ifndef HEAT_TEST
					for (i=1; i<domainHeight_; i++)
						obstacle_grid_[i][1] = NOT_OBSTACLE;
#endif
					break;
		default:
			cerr << "Invalid enumerant in void Navier_Stokes_Solver::add_boundary_condition(Direction dir, float u, float v)" << endl;
			exit(1);
	}

	conditions_.push_front(new UVCondition(cd, new InFlow(u, v, t)));
	set_obstacle();
}

/*****************************************************************************/

void Navier_Stokes_Solver::add_boundary_condition(Direction dir, int start, int finish, float u, float v, float t)
{
	ConditionDirection *cd;
	int i, mn, mx;
	mn = min(start, finish)+1;
	mx = max(start, finish)+1;

	switch (dir) {
		case north:	cd = new NorthCondition(start, finish, domainHeight_);
#ifndef HEAT_TEST
					for (i=mn; i<=mx; i++)
						obstacle_grid_[domainHeight_][i] = NOT_OBSTACLE;
#endif
					break;
		case south:	cd = new SouthCondition(start, finish);
#ifndef HEAT_TEST
					for (i=mn; i<=mx; i++)
						obstacle_grid_[1][i] = NOT_OBSTACLE;
#endif
					break;
		case east:	cd = new EastCondition(start, finish, domainWidth_);
#ifndef HEAT_TEST
					for (i=mn; i<=mx; i++)
						obstacle_grid_[i][domainWidth_] = NOT_OBSTACLE;
#endif
					break;
		case west:	cd = new WestCondition(start, finish);	
#ifndef HEAT_TEST
					for (i=mn; i<=mx; i++)
						obstacle_grid_[i][1] = NOT_OBSTACLE;
#endif
					break;
		default:
			cerr << "invalid enumerant in void Navier_Stokes_Solver::add_boundary_condition(Direction dir, int start, int finish, float u, float v)" << endl;
			exit(1);
	}

	conditions_.push_front(new UVCondition(cd, new InFlow(u, v, t)));
	set_obstacle();
}

/*****************************************************************************/

void Navier_Stokes_Solver::setup_boundary_conditions()
{
	/*add_boundary_condition(north, noSlip);
	add_boundary_condition(south, noSlip);
	add_boundary_condition(east, noSlip);
	add_boundary_condition(west, noSlip);*/
}

/*****************************************************************************/
// GPU program management

void Navier_Stokes_Solver::load_all_programs()
{
	check_errors("before Navier_Stokes_Solver::load_all_programs()");
	const int fp_count = 40;
	string fp_names[fp_count][3] = {
		{"cg_programs/utils.cg", "pass_thru",  "pass-thru"},
		
		{"cg_programs/FG.cg",    "compute_fg",         "compute-fg"},
		{"cg_programs/FG.cg",    "compute_fg_temperature",         "compute-fg-temperature"},

		{"cg_programs/poisson.cg", "rhs",              "rhs"},
		{"cg_programs/poisson.cg", "jacobi_step",      "jacobi-step"},
		
// Hmmm
		{"cg_programs/poisson.cg", "p_residual",         "residual"},
		{"cg_programs/poisson.cg", "p_reduce",           "sum-reduce"},
		{"cg_programs/poisson.cg", "max_reduce",         "max-reduce"},

		{"cg_programs/newuv.cg", "compute_uv",         "compute-uv"},

		//{"cg_programs/boundaries.cg", "uv_pass_thru",       "uv-pass-thru"},
		//{"cg_programs/boundaries.cg", "white",              "white"},
		{"cg_programs/boundaries.cg", "no_slip_south",      "no-slip-south"},
		{"cg_programs/boundaries.cg", "no_slip_west",       "no-slip-west"},
		{"cg_programs/boundaries.cg", "no_slip_east_v",     "no-slip-east-v"},
		{"cg_programs/boundaries.cg", "no_slip_east_u",     "no-slip-east-u"},
		{"cg_programs/boundaries.cg", "no_slip_north_u",    "no-slip-north-u"},
		{"cg_programs/boundaries.cg", "no_slip_north_v",    "no-slip-north-v"},
		{"cg_programs/boundaries.cg", "free_slip_south",    "free-slip-south"},
		{"cg_programs/boundaries.cg", "free_slip_west",     "free-slip-west"},
		{"cg_programs/boundaries.cg", "free_slip_east_v",   "free-slip-east-v"},
		{"cg_programs/boundaries.cg", "free_slip_east_u",   "free-slip-east-u"},
		{"cg_programs/boundaries.cg", "free_slip_north_u",  "free-slip-north-u"},
		{"cg_programs/boundaries.cg", "free_slip_north_v",  "free-slip-north-v"},
		{"cg_programs/boundaries.cg", "outflow_south",      "outflow-south"},
		{"cg_programs/boundaries.cg", "outflow_west",       "outflow-west"},
		{"cg_programs/boundaries.cg", "outflow_east_v",     "outflow-east-v"},
		{"cg_programs/boundaries.cg", "outflow_east_u",     "outflow-east-u"},
		{"cg_programs/boundaries.cg", "outflow_north_u",    "outflow-north-u"},
		{"cg_programs/boundaries.cg", "outflow_north_v",    "outflow-north-v"},
		{"cg_programs/boundaries.cg", "inflow_south",       "inflow-south"},
		{"cg_programs/boundaries.cg", "inflow_west",        "inflow-west"},
		{"cg_programs/boundaries.cg", "inflow_east_v",      "inflow-east-v"},
		{"cg_programs/boundaries.cg", "inflow_east_u",      "inflow-east-u"},
		{"cg_programs/boundaries.cg", "inflow_north_u",     "inflow-north-u"},
		{"cg_programs/boundaries.cg", "inflow_north_v",     "inflow-north-v"},

		{"cg_programs/advection.cg", "ink_advection",  "ink-advection"},
		{"cg_programs/advection.cg", "ink_grid",       "ink-grid"},
		{"cg_programs/advection.cg", "ink_boundary",   "ink-boundary"},
		{"cg_programs/splat.cg",     "ink_splat",      "ink-splat"},
		
		{"cg_programs/obstacle.cg",  "obstacle",       "obstacle"},
		{"cg_programs/obstacle.cg",  "draw_obstacle",  "draw-obstacle"},

		{"cg_programs/temperature.cg", "new_temperature", "new-temperature"}

		//{"cg_programs/ibfv.cg",      "add_noise",      "ibfv-noise"}
	};
	const int vp_count = 4;
	string vp_names[vp_count][3] = {
		{"cg_programs/stencil.cg", "fivestar_stencil",   "fivestar-stencil"},
		{"cg_programs/stencil.cg", "reduce_stencil",     "reduce-stencil"},
		{"cg_programs/stencil.cg", "rhs_stencil",        "rhs-stencil"},
		{"cg_programs/stencil.cg",      "sevenstar_stencil",  "sevenstar-stencil"},
	};
	const int fp_asm_count = 1;
	string fp_asm_names[fp_asm_count][3] = {
		{"cg_programs/max_reduce.fp", "asm_max_reduce",   "asm-max-reduce"},
	};

	CGprofile fp_prof = cgGLGetLatestProfile(CG_GL_FRAGMENT);
	assert(fp_prof >= CG_PROFILE_FP30);

	CGprofile vp_prof = cgGLGetLatestProfile(CG_GL_VERTEX);
	assert(vp_prof >= CG_PROFILE_VP30);

	for (int i=0; i<fp_count; i++)
		CG_Manager::loadProgram(fp_names[i][0], fp_names[i][1], fp_names[i][2], CG_PROFILE_FP30);
	for (int i=0; i<vp_count; i++)
		CG_Manager::loadProgram(vp_names[i][0], vp_names[i][1], vp_names[i][2], CG_PROFILE_VP30);
	for (int i=0; i<fp_asm_count; i++)
		CG_Manager::loadAsmProgram(fp_asm_names[i][0], fp_asm_names[i][1], fp_asm_names[i][2], CG_PROFILE_FP30);

	check_errors("after Navier_Stokes_Solver::load_all_programs()");
}

/*************************************************************************/

void UVCondition::drawBoundary(Render_Buffer& buffer)
{
	check_errors("before UVCondition::drawBoundary(Render_Buffer& buffer)");
	direction_->draw(*type_, buffer);
	check_errors("after UVCondition::drawBoundary(Render_Buffer& buffer)");
}

/*************************************************************************/

void SouthCondition::draw(ConditionType& cond, Render_Buffer& buffer)
{
	check_errors("before SouthCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
	CG_Program p = CG_Manager::getProgram(cond.getName() + "-south");

	p.enable();
	p.setTexture("uv", buffer);
	cond.set_extra_parameters(p);

	glQuadi(getStart(), 0, getFinish()+1, 1);

	p.unsetTexture("uv");
	p.disable();
	check_errors("after SouthCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
}

/*************************************************************************/

void WestCondition::draw(ConditionType& cond, Render_Buffer& buffer)
{
	check_errors("before WestCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
	CG_Program p = CG_Manager::getProgram(cond.getName() + "-west");

	p.enable();
	p.setTexture("uv", buffer);
	cond.set_extra_parameters(p);

	glQuadi(0, getStart(), 1, getFinish()+1);

	p.unsetTexture("uv");
	p.disable();
	check_errors("after WestCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
}

/*************************************************************************/

void NorthCondition::draw(ConditionType& cond, Render_Buffer& buffer)
{
	check_errors("before NorthCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
	CG_Program up = CG_Manager::getProgram(cond.getName() + "-north-u"),
			   vp = CG_Manager::getProgram(cond.getName() + "-north-v");

	glColorMask(true, true, true, true);
	up.enable();
	up.setTexture("uv", buffer);
	cond.set_extra_parameters(up);

	glBegin(GL_LINES);
	glTexCoord2i(getStart(),  height_ - 1);
	glVertex2i  (getStart(),  height_);
	glTexCoord2i(getFinish(), height_ - 1);
	glVertex2i  (getFinish(), height_);
	glEnd();

	up.unsetTexture("uv");
	up.disable();
	
	glColorMask(false, true, false, true);
	vp.enable();
	vp.setTexture("uv", buffer);
	cond.set_extra_parameters(vp);
	glBegin(GL_LINES);
	glTexCoord2i(getStart(),  height_ - 2);
	glVertex2i  (getStart(),  height_ - 1);
	glTexCoord2i(getFinish(), height_ - 2);
	glVertex2i  (getFinish(), height_ - 1);
	glEnd();

	vp.unsetTexture("uv");
	vp.disable();
	glColorMask(true, true, true, true);
	check_errors("after NorthCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
}

/*************************************************************************/

void EastCondition::draw(ConditionType& cond, Render_Buffer& buffer)
{
	check_errors("before EastCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
	CG_Program up = CG_Manager::getProgram(cond.getName() + "-east-u"),
			   vp = CG_Manager::getProgram(cond.getName() + "-east-v");

	glColorMask(true, false, false, true);
	up.enable();
	up.setTexture("uv", buffer);
	cond.set_extra_parameters(up);

	glBegin(GL_LINES);
	glTexCoord2i(width_ - 2, getStart());
	glVertex2i  (width_ - 2, getStart());
	glTexCoord2i(width_ - 2, getFinish());
	glVertex2i  (width_ - 2, getFinish());
	glEnd();

	up.unsetTexture("uv");
	up.disable();

	glColorMask(true, true, true, true);
	vp.enable();
	vp.setTexture("uv", buffer);
	cond.set_extra_parameters(vp);
	glBegin(GL_LINES);
	glTexCoord2i(width_ - 1, getStart());
	glVertex2i  (width_ - 1, getStart());
	glTexCoord2i(width_ - 1, getFinish());
	glVertex2i  (width_ - 1, getFinish());
	glEnd();
	vp.unsetTexture("uv");
	vp.disable();
	glColorMask(true, true, true, true);
	check_errors("after EastCondition::draw(ConditionType& cond, Render_Buffer& buffer)");
}

/*************************************************************************/

void Navier_Stokes_Solver::compute_uv_boundary()
{
	check_errors("before Navier_Stokes_Solver::draw_uv_boundary()");

	CG_Program obstacle = CG_Manager::getProgram("obstacle");

	uv_buffer_->Activate();
	glDrawBuffer(GL_FRONT_AND_BACK);
	
	uv_buffer_->Bind();
	glDisable(GL_DEPTH_TEST);

	obstacle.enable();
	obstacle.setTexture("flag_tex", flag_texture_);
	obstacle.setTexture("uv_tex", *uv_buffer_);

	glQuadi(0, 0, domainWidth_, domainHeight_);

	obstacle.unsetTexture("flag_tex");
	obstacle.unsetTexture("uv_tex");
	obstacle.disable();

	uv_buffer_->Release();
	uv_buffer_->Deactivate();

//	display_buffer(VELOCITY_FIELD);

	uv_buffer_->Activate();
	uv_buffer_->Bind();
	

	glEnable(GL_DEPTH_TEST);
	glClear(GL_DEPTH_BUFFER_BIT);

	list<UVBoundaryCondition*>::iterator itor = conditions_.begin(),
		itor2 = conditions_.end();
	for (; itor != itor2; ++itor)
		(*itor)->drawBoundary(*uv_buffer_);
	
	uv_buffer_->Release();
	uv_buffer_->Deactivate();

	check_errors("after Navier_Stokes_Solver::draw_uv_boundary()");
}

/*************************************************************************/

void Navier_Stokes_Solver::compute_temperature()
{
	check_errors("before Navier_Stokes_Solver::compute_temperature()");
	CG_Program nt = CG_Manager::getProgram("new-temperature");
	uv_buffer_->Activate();
	glDisable(GL_DEPTH_TEST);
	glDrawBuffer(GL_FRONT_AND_BACK);
	glColorMask(false, false, true, false);
	uv_buffer_->Bind();
	nt.enable();
	nt.setTexture("uvt", *uv_buffer_);
	nt.setTexture("obstacle", obstacle_texture_);
	nt.setParameter("dx", dx_);
	nt.setParameter("dy", dy_);
	nt.setParameter("Re", reynoldsNumber_);
	nt.setParameter("Pr", prandtlNumber_);
	nt.setParameter("dt", dt_);
	nt.setParameter("gamma", 0.9f);

	glQuadi(0,0,domainWidth_, domainHeight_);
	glColorMask(true, true, true, true);
	nt.disable();
	uv_buffer_->Release();
	uv_buffer_->Deactivate();
	check_errors("after Navier_Stokes_Solver::compute_temperature()");
}

/*************************************************************************/

void Navier_Stokes_Solver::compute_fg()
{
	check_errors("before Navier_Stokes_Solver::draw_fg()");

	if (heated_) {
		CG_Program compute_fg = CG_Manager::getProgram("compute-fg-temperature"),
			fg_stencil = CG_Manager::getProgram("sevenstar-stencil");

		glDisable(GL_DEPTH_TEST);
		fg_buffer_->Activate();
		glDrawBuffer(GL_FRONT);

		uv_buffer_->Bind();
		fg_stencil.enable();
		fg_stencil.setParameter("mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
		compute_fg.enable();
		compute_fg.setTexture("uv", *uv_buffer_);
		compute_fg.setTexture("flag", flag_texture_);
		compute_fg.setParameter("Re", 1/reynoldsNumber_);
		compute_fg.setParameter("beta", beta_);
		compute_fg.setParameter("dxdy", 1/dx_, 1/(dx_*dx_), 1/dy_, 1/(dy_*dy_));

		compute_fg.setParameter("dt", dx_);
		compute_fg.setParameter("dt", dy_);
		compute_fg.setParameter("dt", dt_);
		compute_fg.setParameter("gamma", 0.9f);
		compute_fg.setParameter("g", gravity_x_, gravity_y_);

		glQuadi(0, 0, domainWidth_, domainHeight_);

		compute_fg.unsetTexture("uv");
		compute_fg.unsetTexture("flag");
		uv_buffer_->Release();
		compute_fg.disable();
		fg_stencil.disable();

		fg_buffer_->Deactivate();


	} else {
		CG_Program compute_fg  = CG_Manager::getProgram("compute-fg"),
				fg_stencil  = CG_Manager::getProgram("sevenstar-stencil");

		glDisable(GL_DEPTH_TEST);
		fg_buffer_->Activate();
		glDrawBuffer(GL_FRONT);

		uv_buffer_->Bind();
		fg_stencil.enable();
		fg_stencil.setParameter("mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
		compute_fg.enable();
		compute_fg.setTexture("uv", *uv_buffer_);
		compute_fg.setTexture("flag", flag_texture_);
		compute_fg.setParameter("Re", 1/reynoldsNumber_);
		compute_fg.setParameter("dxdy", 1/dx_, 1/(dx_*dx_), 1/dy_, 1/(dy_*dy_));

		compute_fg.setParameter("dt", dx_);
		compute_fg.setParameter("dt", dy_);
		compute_fg.setParameter("dt", dt_);
		compute_fg.setParameter("gamma", 0.9f);
		compute_fg.setParameter("g", gravity_x_, gravity_y_);

		glQuadi(0, 0, domainWidth_, domainHeight_);

		compute_fg.unsetTexture("uv");
		compute_fg.unsetTexture("flag");
		uv_buffer_->Release();
		compute_fg.disable();
		fg_stencil.disable();

		fg_buffer_->Deactivate();
	}

	check_errors("after Navier_Stokes_Solver::draw_fg()");
}

/*************************************************************************/

void Navier_Stokes_Solver::solve_pressure_step(GLenum from, GLenum to)
{
	check_errors("before Navier_Stokes_Solver::solve_pressure_step(Render_Buffer& p_from, Render_Buffer& p_to)");

	CG_Program jacobi_step = CG_Manager::getProgram("jacobi-step"),
		jacobi_stencil = CG_Manager::getProgram("fivestar-stencil");

	glDrawBuffer(to);
	p_buffer_->Bind(fromGLenum(from));

	fg_buffer_->Bind();

	glDisable(GL_DEPTH_TEST);

	jacobi_stencil.enable();
	jacobi_stencil.setParameter("mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	jacobi_step.enable();
	jacobi_step.setParameter("dxdy",			   1.0f/(dx_*dx_), 1.0f/(dy_*dy_), -1.0f);
	jacobi_step.setParameter("one_over_dxdt_dydt", 1.0f/(dx_*dt_), 1.0f/(dy_*dt_), 0);

	jacobi_step.setTexture("fg",    *fg_buffer_);
	jacobi_step.setTexture("p_old", *p_buffer_);
	jacobi_step.setTexture("neighbors",  neighbors_texture_);

	glQuadi(1, 1, domainWidth_-1, domainHeight_-1);

	jacobi_step.unsetTexture("p_old");
	jacobi_step.unsetTexture("fg");
	jacobi_step.unsetTexture("neighbors");
	jacobi_step.disable();
	jacobi_stencil.disable();
    
    fg_buffer_->Release();

	p_buffer_->Release();

    check_errors("after Navier_Stokes_Solver::solve_pressure_step(Render_Buffer& p_from, Render_Buffer& p_to)");
}

/*****************************************************************************/

float Navier_Stokes_Solver::get_residual(Render_Buffer &p, Surface_Type surface)
{
	check_errors("before float Navier_Stokes_Solver::get_residual(Render_Buffer &p)");

	CG_Program 
		residual = CG_Manager::getProgram("residual"),
		five_star_stencil = CG_Manager::getProgram("fivestar-stencil"),
		sum_reduce = CG_Manager::getProgram("sum-reduce");

	r_buffer_->Activate();
	glClearColor(0,0,0,0);
	glClear(GL_COLOR_BUFFER_BIT);
	glDrawBuffer(GL_FRONT);

	p.Bind(surface);
	fg_buffer_->Bind();

    five_star_stencil.enable();
	five_star_stencil.setParameter("mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

	residual.enable();
	residual.setParameter("one_over_dxdt_dydt", 1.0f/(dx_*dt_), 1.0f/(dy_*dt_), 0);
    residual.setParameter("dxdy",				1.0f/(dx_*dx_), 1.0f/(dy_*dy_), -1);

	residual.setTexture("fg",        *fg_buffer_);
	residual.setTexture("p",         p);
	residual.setTexture("obstacle",  obstacle_texture_);
	residual.setTexture("neighbors", neighbors_texture_);

	glQuadi(1, 1, domainWidth_-1, domainHeight_-1);

	residual.unsetTexture("p");
	residual.unsetTexture("fg");
	residual.unsetTexture("obstacle");
	residual.unsetTexture("neighbors");
	residual.disable();
    five_star_stencil.disable();

    fg_buffer_->Release();

    p.Release();

	float result[4];
	reduce("sum-reduce", result);

	r_buffer_->Deactivate();

	float residual_v = sqrt(result[0]/(float)((domainHeight_-2)*(domainWidth_-2)));

	check_errors("after float Navier_Stokes_Solver::get_residual(Render_Buffer &p)");
	return residual_v;
}

/*****************************************************************************/
// Navier_Stokes_Solver::reduce assumes that the reduction domain is
// already at the appropriate position in r_buffer_
// We also assume that r_buffer_ is currently activated.

void Navier_Stokes_Solver::reduce(char *program_name, float result[4])
{
#if 1
  CG_Program reduce_kernel = CG_Manager::getProgram(program_name);
  CG_Program reduce_stencil = CG_Manager::getProgram("reduce-stencil");

  int w = r_buffer_->GetWidth(),
    size = w >> 1,
    h = 0;


  GLenum 
	b_from = GL_FRONT,
	b_to   = GL_BACK;

  glDrawBuffer(b_to);
  reduce_stencil.enable();
  reduce_stencil.setParameter("mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
  reduce_kernel.enable();
  r_buffer_->Bind(fromGLenum(b_from));
  reduce_kernel.setTexture("tex", *r_buffer_);
  
  while (size) {
    // set a fence
    glSetFenceNV(fence_, GL_ALL_COMPLETED_NV);
	//glBegin(GL_QUADS);
	//glTexCoord2i(
	//glEnd();
	glQuadii(0, 0, w<<1, w<<1, 0, 0, w, w);
	
    // To partially flush the pipeline on nv30 and nv40s,
    // throw a fence, rebind the pbuffers, and don't ask.
    // Thanks, Cliff Woolley :D
    glFinishFenceNV(fence_);
	reduce_kernel.unsetTexture("tex");
    r_buffer_->Release();
	reduce_kernel.disable();
	reduce_stencil.disable();

	//r_buffer_->Deactivate();
	swap(b_to, b_from);

	//r_buffer_->display_buffer(1, GL_FRONT);
	//r_buffer_->display_buffer(1, GL_BACK);

	//r_buffer_->Activate();
	glDrawBuffer(b_to);
	glClear(GL_COLOR_BUFFER_BIT);
    r_buffer_->Bind(fromGLenum(b_from));

	reduce_stencil.enable();
	reduce_stencil.setParameter("mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	reduce_kernel.enable();
	reduce_kernel.setTexture("tex", *r_buffer_);


    h+=size;
    size >>= 1;
  }
  
  reduce_kernel.unsetTexture("tex");
  r_buffer_->Release();
  reduce_kernel.disable(); 
  reduce_stencil.disable();

  glReadBuffer(b_from);
  glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, result);
#else
	float *v = new float[4 * r_buffer_->GetWidth() * r_buffer_->GetHeight()];
	
	glReadBuffer(GL_FRONT);
	glReadPixels(0, 0, r_buffer_->GetWidth(), r_buffer_->GetHeight(), GL_RGBA, GL_FLOAT, v);
	float *tv = v;
	if (string(program_name) == string("sum-reduce")) {
		result[0] = 0;
		result[1] = 0;
		result[2] = 0;
		result[3] = 0;
		for (int i=0; i<r_buffer_->GetWidth() * r_buffer_->GetHeight(); ++i) {
			result[0] += *tv++;
			result[1] += *tv++;
			result[2] += *tv++;
			result[3] += *tv++;
		}
	} else {
		result[0] = tv[0];
		result[1] = tv[1];
		result[2] = tv[2];
		result[3] = tv[3];
		for (int i=0; i<r_buffer_->GetWidth() * r_buffer_->GetHeight(); ++i) {
			result[0] = max(*tv++, result[0]);
			result[1] = max(*tv++, result[1]);
			result[2] = max(*tv++, result[2]);
			result[3] = max(*tv++, result[3]);
		}
	}
	delete v;
#endif
	glDrawBuffer(GL_FRONT_AND_BACK);
}

/*****************************************************************************/

void Navier_Stokes_Solver::get_max_velocity(float &u, float &v)
{
	check_errors("before void Navier_Stokes_Solver::get_max_velocity(float &u, float &v)");

	CG_Program uv_passthru = CG_Manager::getProgram("pass-thru");

	r_buffer_->Activate();
	glClear(GL_COLOR_BUFFER_BIT);

	uv_buffer_->Bind();
	uv_passthru.enable();
	uv_passthru.setTexture("tex", *uv_buffer_);
	glQuadi(0, 0, domainWidth_, domainHeight_);
	uv_buffer_->Release();
	uv_passthru.disable();

	float result[4];
	reduce("max-reduce", result);

	u = result[0];
	v = result[1];
	max_u_ = u;
	max_v_ = v;
	r_buffer_->Deactivate();

	check_errors("after void Navier_Stokes_Solver::get_max_velocity(float &u, float &v)");
}

/*****************************************************************************/

void Navier_Stokes_Solver::adjust_time_step()
{
	static int counter = steps_per_time_adjustment_ - 1;
	counter++;
	if (counter == steps_per_time_adjustment_) {
		counter = 0;
		float u, v;
		get_max_velocity(u, v);
		dt_ = min(min((reynoldsNumber_ / 2.0f) * (1.0f/(1.0f/(dx_*dx_) + 1.0f/(dy_*dy_))),
							dx_/u), dy_/v);
		if (heated_) {
			dt_ = min(dt_, (reynoldsNumber_ * prandtlNumber_ / 2.0f) * (1.0f/(1.0f/(dx_*dx_) + 1.0f/(dy_*dy_))));
			//dt_ = min(dt_, 0.1);
		}

		dt_ = tau_ * dt_;
#ifdef _DEBUG
		cout << "New dt: " << dt_ << endl;
#endif
	}
}

/*****************************************************************************/

void Navier_Stokes_Solver::solve_pressure()
{
	static int jacobi_steps = 32;
	static int counter = 0;
	static bool turbo_mode = false;
	static int old_timestep_counter;

	int maxiter = 64;
	int turbo_cycle = 128;

	check_errors("before Navier_Stokes_Solver::solve_pressure()");
	float eps = 1e-2f;

	while (1) {
		p_buffer_->Activate();
		for (int i=0; i<jacobi_steps; i++) {
			solve_pressure_step(pressure_s1, pressure_s2);
			swap(pressure_s1, pressure_s2);
		}
		p_buffer_->Deactivate();

#ifdef TURBO_MODE_ENABLED
		if (turbo_mode) {
			counter++;
#ifdef _DEBUG
			cout << ".";
			cout.flush();
#endif
			if (counter == turbo_cycle) {
				counter = 0;
#ifdef _DEBUG
				cout << "Checking turbo residual" << endl;
#endif
				float f = get_residual(*p_buffer_, fromGLenum(pressure_s1));
				if (f > eps) {
					turbo_mode = false;
#ifdef _DEBUG
					cout << "Leaving turbo" << endl;
					steps_per_time_adjustment_ = old_timestep_counter;
#endif
				} else {
#ifdef _DEBUG
					cout << "Turbo still good" << endl;
#endif
				}
			}
			break;
		} else {
			float f = get_residual(*p_buffer_, fromGLenum(pressure_s1));
			if (f == 0)
				cout << "Sanity check: residual is zero. Something may have gone wrong.." << endl;

			if (f > eps) {
				jacobi_steps *= 2;
				counter = 0;
				if (jacobi_steps > maxiter) {
					jacobi_steps = maxiter;
#ifdef _DEBUG
					cerr << "Leaving solve_pressure with over-epsilon r: " << f << endl;
#endif
					break;
				}
			} else {
				jacobi_steps /= 2;
				if (jacobi_steps < 4) {
					jacobi_steps = 4;
					counter++;
					if (counter == 128) {
						// achieved steady state?
#ifdef _DEBUG
						cout << "Entering turbo mode!" << endl;
#endif
						cout << "Stable!" << endl;
						turbo_mode = true;
						counter = 0;
						old_timestep_counter = steps_per_time_adjustment_;
						steps_per_time_adjustment_ = 512;
					}
				}
#ifdef _DEBUG
				cout << "Ok, steps = " << jacobi_steps << ", r: " << f << endl;
#endif
				break;
			}
		}
#else
		float f = get_residual(*p_buffer_, fromGLenum(pressure_s1));
		if (f == 0)
				cout << "Sanity check: residual is zero. Something may have gone wrong.." << endl;

		if (f > eps) {
			jacobi_steps *= 2;
			counter = 0;
			if (jacobi_steps > maxiter) {
				jacobi_steps = maxiter;
#ifdef _DEBUG
				cerr << "Leaving solve_pressure with over-epsilon r: " << f << endl;
#endif
				break;
			}
		} else {
			jacobi_steps /= 2;
			if (jacobi_steps < 4)
				jacobi_steps = 4;
#ifdef _DEBUG
				cout << "Ok, steps = " << jacobi_steps << ", r: " << f << endl;
#endif
				break;
		}
#endif
	}

	check_errors("after Navier_Stokes_Solver::solve_pressure()");
}

/*************************************************************************/

void Navier_Stokes_Solver::compute_new_uv()
{
	check_errors("before Navier_Stokes_Solver::compute_new_uv()");

	CG_Program compute_uv = CG_Manager::getProgram("compute-uv");
	
	uv_buffer_->Activate();
	glColorMask(true, true, false, true);
	glDisable(GL_DEPTH_TEST);
	
	glDrawBuffer(GL_FRONT_AND_BACK);

	fg_buffer_->Bind();
	p_buffer_->Bind();
	compute_uv.enable();
	compute_uv.setParameter("dt", dt_);
	compute_uv.setParameter("dx", dx_);
	compute_uv.setParameter("dy", dy_);
	compute_uv.setTexture("fg", *fg_buffer_);
	compute_uv.setTexture("p", *p_buffer_);

	glQuadi(1, 1, domainWidth_ -1, domainHeight_-1);

	compute_uv.unsetTexture("fg");
	compute_uv.unsetTexture("p");
	compute_uv.disable();
	fg_buffer_->Release();
	p_buffer_->Release();
	glColorMask(true, true, true, true);

	uv_buffer_->Deactivate();


	check_errors("after Navier_Stokes_Solver::compute_new_uv()");
}

void Navier_Stokes_Solver::test_buffer()
{
	check_errors("before Navier_Stokes_Solver::test_buffer()");

	CG_Program checkered = CG_Manager::getProgram("checkered");
	uv_buffer_->Activate();
	checkered.enable();
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f(1.0f, 1.0f, 1.0f);

	int w = uv_buffer_->GetWidth(),
		h = uv_buffer_->GetHeight();

	glBegin(GL_QUADS);

	glTexCoord2i(0,0);
	glVertex2i  (0,0);

	glTexCoord2i(1,0);
	glVertex2i  (w,0);

	glTexCoord2i(1,1);
	glVertex2i  (w,h);

	glTexCoord2i(0,1);
	glVertex2i  (0,h);

	glEnd();

	checkered.disable();
	uv_buffer_->Deactivate();

	check_errors("after Navier_Stokes_Solver::test_buffer()");
}

/*************************************************************************/

void Navier_Stokes_Solver::splat_ink()
{
	check_errors("before Navier_Stokes_Solver::splat_ink()");

	CG_Program splat = CG_Manager::getProgram("ink-splat");

	ink_buffer_->Activate();
	ink_buffer_->Bind(fromGLenum(ink_from));
	glDrawBuffer(ink_from);
	splat.enable();
	splat.setTexture("base", *ink_buffer_);

	vector<Splat>::iterator b = splats_.begin(),
			e = splats_.end();

	for (; b != e; ++b) {
		splat.setParameter("splatcolor", 
							b->color_[0],
							b->color_[1],
							b->color_[2],
							b->color_[3]);
		splat.setParameter("radius", 1.0);
		
		float r = b->radius_ * (float) inkSuperSampling_;
		float x1 = b->center_[0] * (float) inkSuperSampling_ - r, 
			y1 = b->center_[1] * (float) inkSuperSampling_ - r,
			x2 = b->center_[0] * (float) inkSuperSampling_ + r, 
			y2 = b->center_[1] * (float) inkSuperSampling_ + r;

		glBegin(GL_QUADS);
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, x1, y1);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, -5, -5);
		glVertex2f(x1, y1);
	
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, x2, y1);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 5, -5);
		glVertex2f(x2, y1);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, x2, y2);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 5, 5);
		glVertex2f(x2, y2);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, x1, y2);
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, -5, 5);
		glVertex2f(x1, y2);
		glEnd();
	}
	
	splat.unsetTexture("base");
	splat.disable();
	
	ink_buffer_->Release();
	
	ink_buffer_->Deactivate();

	check_errors("after Navier_Stokes_Solver::splat_ink()");
}

/*************************************************************************/

void Navier_Stokes_Solver::clear_ink()
{
	ink_buffer_->Activate();
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	ink_buffer_->Deactivate();
}

/*************************************************************************/

void Navier_Stokes_Solver::ink_grid()
{
	check_errors("before void Navier_Stokes_Solver::ink_grid()");

	CG_Program ink_grid = CG_Manager::getProgram("ink-grid");

	ink_buffer_->Activate();
	ink_grid.enable();
	ink_grid.setParameter("color", 1.0, 1.0, 1.0, 1.0);
	glDrawBuffer(GL_FRONT_AND_BACK);

	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex2i  (0,0);
	glTexCoord2f(inkSuperSampling_*domainWidth_/gridScale_,0);
	glVertex2i  (inkSuperSampling_*domainWidth_,0);
	glTexCoord2f(inkSuperSampling_*domainWidth_/gridScale_,inkSuperSampling_*domainHeight_/gridScale_);
	glVertex2i  (inkSuperSampling_*domainWidth_,inkSuperSampling_*domainHeight_);
	glTexCoord2f(0,inkSuperSampling_*domainHeight_/gridScale_);
	glVertex2i  (0,inkSuperSampling_*domainHeight_);
	glEnd();
	
	ink_grid.disable();
	ink_buffer_->Deactivate();

	check_errors("after void Navier_Stokes_Solver::ink_grid()");
}

/*************************************************************************/

void Navier_Stokes_Solver::ibfv()
{
	check_errors("before Navier_Stokes_Solver::ibfv()");

	CG_Program noise = CG_Manager::getProgram("ibfv-noise");

	/*ink_buffer_->Activate();
	glDrawBuffer(ink_to);
	ink_buffer_->Bind(fromGLenum(ink_from));

	noise.enable();
	noise.setTexture("noise", noise_texture_);
	noise.setTexture("ink",   *ink_buffer_);
	noise.setParameter("alpha", 0.01f);
	glQuadii(0, 0, domainWidth_, domainHeight_,
 		     0, 0, inkSuperSampling_*domainWidth_, inkSuperSampling_ * domainHeight_);
	noise.disable();

	ink_buffer_->Release();
	ink_buffer_->Deactivate();

	//swap(ink_from, ink_to);*/

	check_errors("after Navier_Stokes_Solver::ibfv()");
}

/*************************************************************************/

void Navier_Stokes_Solver::advect_ink()
{
	check_errors("before void Navier_Stokes_Solver::advect_ink()");

	CG_Program advect_ink = CG_Manager::getProgram("ink-advection"),
			 ink_boundary = CG_Manager::getProgram("ink-boundary");

	ink_buffer_->Activate();
	glDrawBuffer(ink_to);
	ink_buffer_->Bind(fromGLenum(ink_from));
	ink_boundary.enable();
	ink_boundary.setTexture("ink", *ink_buffer_);
	ink_boundary.setTexture("obstacle", obstacle_texture_);
	ink_boundary.setTexture("neighbors", neighbors_texture_);

	glQuadi(0, 0, domainWidth_, domainHeight_);

	ink_boundary.unsetTexture("ink");
	ink_boundary.unsetTexture("obstacle");
	ink_boundary.unsetTexture("neighbors");
	ink_boundary.disable();

	uv_buffer_->Bind();
	advect_ink.enable();
	advect_ink.setParameter("timestep", dt_);
	advect_ink.setParameter("dissipation", 1.00f);
	advect_ink.setParameter("dxdy", dx_, dy_);
	advect_ink.setParameter("ssScale", (float)inkSuperSampling_);
	advect_ink.setParameter("turboInk", (float)turboInk_);
	advect_ink.setTexture("ink", *ink_buffer_);
	advect_ink.setTexture("uv", *uv_buffer_);
	{
		int i = (int)fabs(totalTime_*ibfvTimeScale_);
		advect_ink.setTexture("noise1", noise_texture_[i % noise_texture_count_]);
		advect_ink.setTexture("noise2", noise_texture_[(i+1) % noise_texture_count_]);
	}
	if (ibfv_)
		advect_ink.setParameter("noise_alpha", currentIbfvAlpha_);
	else
		advect_ink.setParameter("noise_alpha", 0.00f);
	advect_ink.setParameter("noise_modulate", totalTime_*ibfvTimeScale_);
	
	glQuadii(0, 0, domainWidth_, domainHeight_,
 		     0, 0, inkSuperSampling_*domainWidth_, inkSuperSampling_ * domainHeight_);

	uv_buffer_->Release();
	advect_ink.unsetTexture("ink");
	advect_ink.unsetTexture("uv");
	advect_ink.disable();


	ink_buffer_->Release();
	
	ink_buffer_->Deactivate();

	swap(ink_from, ink_to);

	check_errors("after void Navier_Stokes_Solver::advect_ink()");
}

/*************************************************************************/

void Navier_Stokes_Solver::draw_obstacle()
{
	check_errors("before void Navier_Stokes_Solver::draw_obstacle()");

	CG_Program obstacle = CG_Manager::getProgram("draw-obstacle");

	glEnable(GL_TEXTURE_RECTANGLE_NV);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, obstacle_texture_);
	glColorMask(true, true, true, true);
	obstacle.enable();
	obstacle.setTexture("obstacle", obstacle_texture_);

	glQuadii(0, 0, domainWidth_, domainHeight_, 0, 0, 1, 1);

	obstacle.unsetTexture("obstacle");
	obstacle.disable();

	check_errors("after void Navier_Stokes_Solver::draw_obstacle()");
}

/*************************************************************************/

void Navier_Stokes_Solver::display()
{
	check_errors("before Navier_Stokes_Solver::display()");

	glMatrixMode(GL_MODELVIEW);
	
	switch (displayBuffer_) {
		case VELOCITY_FIELD_X:
			uv_buffer_->draw_on_frame_buffer(0, SURFACE_FRONT);
			break;
		case VELOCITY_FIELD_Y:
			uv_buffer_->draw_on_frame_buffer(1, SURFACE_FRONT);
			break;
		case INK_FIELD:
			ink_buffer_->draw_on_frame_buffer(-1, SURFACE_FRONT, lerp_);
			draw_obstacle();
			break;
		case FLUID_DOMAIN:
			glClearColor(0.0, 0.0, 0.0, 0.0);
			glClear(GL_COLOR_BUFFER_BIT);
			draw_obstacle();
			break;
	}

	check_errors("after Navier_Stokes_Solver::display()");
}

/*************************************************************************/

void Navier_Stokes_Solver::test_draw()
{
	check_errors("before Navier_Stokes_Solver::test_draw()");

	// Clear Screen
	glClear(GL_COLOR_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0f, 3.0f, 0.0f, 3.0f, -1.0f, 1.0f);

	glMatrixMode(GL_MODELVIEW);
	
	uv_buffer_->draw_on_frame_buffer(0, SURFACE_FRONT);
	glPushMatrix();
	glTranslatef(1.0f, 0.0f, 0.0f);
	uv_buffer_->draw_on_frame_buffer(1, SURFACE_FRONT);
	glPopMatrix();
	glPushMatrix();
	glTranslatef(1.0f, 1.0f, 0.0f);
	ink_buffer_->draw_on_frame_buffer(-1, SURFACE_FRONT);
	glPopMatrix();

	check_errors("after Navier_Stokes_Solver::test_draw()");
}


/*************************************************************************/

void Navier_Stokes_Solver::display_buffer(int buffer)
{
	switch (buffer)	{
	case VELOCITY_FIELD: cout << "Velocity Field:" << endl;
			uv_buffer_->display_buffer(3, GL_FRONT);
			uv_buffer_->display_buffer(3, GL_BACK);
			return;
	case FG_FIELD: cout << "Tentative Velocity Field (FG):" << endl;
			fg_buffer_->display_buffer(2, GL_FRONT);
			fg_buffer_->display_buffer(2, GL_BACK);
			return;
	case PRESSURE_FIELD: cout << "Pressure Field:" << endl;
		    p_buffer_->display_buffer(1, GL_FRONT);
			return;
	case INK_FIELD: cout << "Ink Field:" << endl;
			ink_buffer_->display_buffer(1, GL_FRONT);
			return;
	}
}

/*************************************************************************/
/* Constructing from File */

Navier_Stokes_Solver::Navier_Stokes_Solver() {
	inkTime_ = pressureTime_ = timestepTime_ = fgTime_ = boundaryTime_ = newuvTime_ = 0;
	inkQuality_ = iqFloating;
}

void Navier_Stokes_Solver::load_from_file(char *confFile)
{
	ifstream in(confFile);

	bool created = false;

	domainWidth_ = 50;
	domainHeight_ = 50;
	reynoldsNumber_ = 100.0;
	inkSuperSampling_ = 1;
	turboInk_ = 1;

	string token;

	while (in >> token) {
		if (token == "inkquality") {
			string str;
			in >> str;
			inkQuality_ = inkQualityFromString(str);
		} else if (token == "width") {
			in >> domainWidth_;
			domainWidth_ += 2;
		} else if (token == "height") {
			in >> domainHeight_;
			domainHeight_ += 2;
		} else if (token == "reynolds")
			in >> reynoldsNumber_;
		else if (token == "heated") {
			string str;
			in >> str;
			if (str == "true")
				heated_ = true;
			else
				heated_ = false;
		} else if (token == "prandtl")
			in >> prandtlNumber_;
		else if (token == "beta")
			in >> beta_;
		else if (token == "inksupersampling")
			in >> inkSuperSampling_;
		else if (token == "xlength")
			in >> xLength_;
		else if (token == "ylength")
			in >> yLength_;
		else if (token == "turboink")
			in >> turboInk_;
		else if (token == "gravity")
			in >> gravity_x_ >> gravity_y_;
		else if (token == "boundary") {
			if (!created) {
				init_solver();
				created = true;
			}
			string bType;
			in >> bType;
			if (bType == "inflow") {
				string dir;
				int from, to;
				float u, v, t;
				in >> dir >> from >> to >> u >> v >> t;
				add_boundary_condition(directionFromString(dir), from, to, u, v, t);
			} else if (bType == "outflow") {
				string dir;
				int from, to;
				in >> dir >> from >> to;
				add_boundary_condition(directionFromString(dir), outFlow, from, to);
			} else if (bType == "freeslip") {
				string dir;
				int from, to;
				in >> dir >> from >> to;
				add_boundary_condition(directionFromString(dir), freeSlip, from, to);
			}
		} else if (token == "splat") {
			if (!created) {
				init_solver();
				created = true;
			}
			string color, size;
			float x, y;
			in >> color >> size >> x >> y;
			splat_choose(colorSizeFromString(color));
			splat_choose(colorSizeFromString(size));
			Splat s = get_custom_splat();
			s.center_[0] = x;
			s.center_[1] = y;
			add_splat(s);
		} else if (token == "gridscale") {
			if (!created) {
				init_solver();
				created = true;
			}
			float s;
			in >> s;
			set_grid_scale(s);
		} else if (token == "steps") {
			if (!created) {
				init_solver();
				created = true;
			}
			float s;
			in >> s;
			steps_per_time_adjustment_ = (int) s;
		} else if (token == "tau") {
			if (!created) {
				init_solver();
				created = true;
			}
			float s;
			in >> s;
			tau_ = s;
		} else if (token == "lerp") {
			if (!created) {
				init_solver();
				created = true;
			}
			string s;
			in >> s;
			if (s == "true")
				lerp_ = true;
			else
				lerp_ = false;
		} else if (token == "obstacle") {
			if (!created) {
				init_solver();
				created = true;
			}
			string s, t;
			for (int i=0; i<domainHeight_; i++) {
				in >> t;
				s = t + s;
			}
			set_obstacle_grid(s.c_str());
			set_obstacle();
		} else if (token == "dump") {
			if (!created) {
				init_solver();
				created = true;
			}
			in >> sequence_name_;
			in >> frame_skip_;
			dump_ = true;
		} else if (token == "ibfv") {
			if (!created) {
				init_solver();
				created = true;
			}
			ibfv_ = true;
			in >> ibfvTimeScale_;
			in >> ibfvAlpha_;
		} else {
			cerr << "Unrecognized token: " << token << endl;
			exit(1);
		}

	}
}

void Navier_Stokes_Solver::reset_stats()
{
	inkTime_ = pressureTime_ = timestepTime_ = fgTime_ = boundaryTime_ = newuvTime_ = totalTime_ = 0;
}

void Navier_Stokes_Solver::show_stats()
{
	float sum = inkTime_ + pressureTime_ + timestepTime_ + fgTime_ + boundaryTime_ + newuvTime_;
	cout << "Total time: " << sum << endl;
	cout << "Time spent in ink advection: " << inkTime_ / sum * 100.0f << "%" << endl;
	cout << "Time spent solving Poisson: " << pressureTime_ / sum * 100.0f << "%" << endl;
	cout << "Time spent adjusting timestep: " << timestepTime_ / sum * 100.0f << "%" << endl;
	cout << "Time spent calculating FG: " << fgTime_ / sum * 100.0f << "%" << endl;
	cout << "Time spent calculating new velocity field: " << newuvTime_ / sum * 100.0f << "%" << endl;
	cout << "Time spent setting boundaries: " << boundaryTime_ / sum * 100.0f << "%" << endl;
}

void Navier_Stokes_Solver::step()
{
	timer.Start();
	timer.Reset();
//	display_buffer(VELOCITY_FIELD);
	compute_uv_boundary();
//	display_buffer(VELOCITY_FIELD);
	float boundary = timer.GetTime();
	if (heated_) {
		compute_temperature();
	}
//	display_buffer(VELOCITY_FIELD);
	if (advection_)
		advect_ink();
//	display_buffer(VELOCITY_FIELD);
	currentIbfvAlpha_ = (currentIbfvAlpha_ + ibfvAlpha_) / 2.0f; // this is to initialize the ink quickly
	splat_ink();

	float ink = timer.GetTime();
	compute_fg();
//	display_buffer(FG_FIELD);
	float fg = timer.GetTime();
	solve_pressure();
	float pressure = timer.GetTime();
	compute_new_uv();
//	display_buffer(VELOCITY_FIELD);
	float newuv = timer.GetTime();
	adjust_time_step();
	float timestep = timer.GetTime();

	timestepTime_ += timestep - newuv;
	newuvTime_ += newuv - pressure;
	pressureTime_ += pressure - fg;
	fgTime_ += fg - ink;
	inkTime_ += ink - boundary;
	boundaryTime_ += boundary;
	timer.Stop();

	if (dump_) {
		dump();
	}

	totalTime_ += dt_;
}

void Navier_Stokes_Solver::dump()
{
	frame_counter_++;
	if (frame_counter_ != frame_skip_)
		return;

	frame_counter_ = 0;
	char v[100];
	itoa(total_frame_counter_++, v, 10);

	string frame_name(sequence_name_ + string(v) + string(".dmp"));
	cout << "Dump " << frame_name << endl;
	ofstream frame(frame_name.c_str());
	ink_buffer_->dump_to_stream(frame, 4);	
}
