#ifndef NAVIER_STOKES_H
#define NAVIER_STOKES_H

#include "render_buffer.h"
#include "cg_manager.h"
#include "stopwatch.hpp"

#include <string>
#include <vector>
#include <list>
#include <cmath>

using namespace std;

/**************************************************************************/
// Splats

class Splat
{
public:
	float radius_;
	float center_[2];
	float color_[4];
};

enum SPLAT_MENU_CHOICES {
	SPLAT_RED,
	SPLAT_GREEN,
	SPLAT_BLUE,
	SPLAT_CYAN,
	SPLAT_MAGENTA,
	SPLAT_YELLOW,
	SPLAT_BLACK,
	SPLAT_WHITE,
	SPLAT_COLORS,

	SPLAT_SMALL,
	SPLAT_MEDIUM,
	SPLAT_LARGE,
	SPLAT_HUGE,
};

/**************************************************************************/

enum InkQuality { iqFixed, iqFloating };
enum Direction { north, south, east, west };
enum Type { noSlip, freeSlip, inFlow, outFlow };
enum Field { 
	VELOCITY_FIELD, 
	VELOCITY_FIELD_X, 
	VELOCITY_FIELD_Y, 
	FG_FIELD,
	PRESSURE_FIELD,
	INK_FIELD,
	FLUID_DOMAIN,
	NEIGHBORS,
};
enum OBSTACLE_FLAG {
	EN, CNE, EE, CSE, ES, CSW, EW, CNW, FS, FSW, FW, FLUID, WALL, OBSTACLE, NOT_OBSTACLE
}; 

/**************************************************************************/

class UVBoundaryCondition
{
public:
	virtual void drawBoundary(Render_Buffer &pbuffer) = 0;
};

/**************************************************************************/

class ConditionType
{
public:
	Type type_;
	string name_;

	ConditionType(Type type): type_(type) {};
	virtual void set_extra_parameters(CG_Program& prog) = 0;
	const string& getName() { return name_; }
};

/**************************************************************************/

class ConditionDirection
{
private:
	int start_, finish_;

public:
	Direction direction_;

	ConditionDirection(int start, int finish):
	  start_(start), finish_(finish) {};
	virtual void draw(ConditionType &type, Render_Buffer& buffer) = 0;

	int getStart() { return start_; };
	int getFinish() { return finish_; };
};

/**************************************************************************/

class SouthCondition: public ConditionDirection
{
public:
	SouthCondition(int start, int finish):
	  ConditionDirection(start, finish) {
	  direction_ = south;
    };

	virtual void draw(ConditionType &type, Render_Buffer& buffer);
};

class NorthCondition: public ConditionDirection
{
public:
	int height_;
	NorthCondition(int start, int finish, int height):
	  ConditionDirection(start, finish),
	  height_(height) {
		direction_ = north;
	  };

    virtual void draw(ConditionType &type, Render_Buffer& buffer);
};

class EastCondition: public ConditionDirection
{
public:
	int width_;
	EastCondition(int start, int finish, int width):
	  ConditionDirection(start, finish),
	  width_(width) {
		direction_ = east;
	  };

    virtual void draw(ConditionType &type, Render_Buffer& buffer);
};

class WestCondition: public ConditionDirection
{
public:
	WestCondition(int start, int finish):
	  ConditionDirection(start, finish) {
		direction_ = west;
	  };

    virtual void draw(ConditionType &type, Render_Buffer& buffer);
};

/*************************************************************************/

class NoSlip: public ConditionType
{
public:
	virtual void set_extra_parameters(CG_Program& prog) {};
	NoSlip(): ConditionType(noSlip) { name_ = "no-slip"; };
};

class FreeSlip: public ConditionType
{
public:
	virtual void set_extra_parameters(CG_Program& prog) {};
	FreeSlip(): ConditionType(freeSlip) { name_ = "free-slip"; };
};

class OutFlow: public ConditionType
{
public:
	virtual void set_extra_parameters(CG_Program& prog) {};
	OutFlow(): ConditionType(outFlow) { name_ = "outflow"; };
};

class InFlow: public ConditionType
{
protected:
	float fu_, fv_, ft_;
public:
	virtual void set_extra_parameters(CG_Program& prog)
	{
		prog.setParameter("boundary_uvt", fu_, fv_, ft_);
	};

	InFlow(float fu, float fv, float ft): 
		ConditionType(inFlow), fu_(fu), fv_(fv), ft_(ft) { name_ = "inflow"; };
};

/*************************************************************************/

class UVCondition: public UVBoundaryCondition
{
private:
	ConditionDirection *direction_;
	ConditionType *type_;

public:
	virtual void drawBoundary(Render_Buffer &pbuffer);

	UVCondition(ConditionDirection *direction,
				ConditionType *type):
	    direction_(direction),
		type_(type),
		UVBoundaryCondition() {};
};

/**************************************************************************/

class Navier_Stokes_Solver
{
private:
	int domainWidth_, domainHeight_;
	float reynoldsNumber_;
	float dx_, dy_, dt_;
	float xLength_, yLength_;
	float prandtlNumber_, beta_;

	float turboInk_;
	int inkSuperSampling_;

	float inkTime_, pressureTime_, fgTime_, boundaryTime_, timestepTime_, newuvTime_;

	float gravity_x_, gravity_y_;

	float gridScale_;
	float tau_;
	float totalTime_;
	int displayBuffer_;
	int steps_per_time_adjustment_;

	float max_u_, max_v_;

	vector<Splat> splats_;
	list<UVBoundaryCondition*> conditions_;
	GLuint flag_texture_;
	GLuint obstacle_texture_;
	GLuint neighbors_texture_;

	// IBFV! Yay! :)
	static const int noise_texture_count_ = 16;
	GLuint noise_texture_[noise_texture_count_];
	bool ibfv_;
	bool advection_;
	InkQuality inkQuality_;
	float ibfvTimeScale_;
	float ibfvAlpha_;
	float currentIbfvAlpha_;
	
	GLuint fence_;
	bool lerp_;

	string sequence_name_;
	int frame_skip_;
	int frame_counter_;
	int total_frame_counter_;
	bool dump_;
	bool heated_;

	float splatColor[4];
	float splatRadius;
	Splat get_custom_splat();
	void splat_choose(int choice);
	
	void init_buffers();
	void init_solver();

	void draw_obstacle();

	void test_draw();
	void test_buffer();
	void init_ibfv();

	vector<vector<unsigned char> > obstacle_grid_;

public:
	Render_Buffer *fg_buffer_,  *uv_buffer_,
				  *p_buffer_,   *r_buffer_, 
				  *ink_buffer_, *flag_buffer_;

	GLenum pressure_s1, pressure_s2;
	GLenum ink_to, ink_from;

	Stopwatch timer;

	void load_all_programs();
	void setup_boundary_conditions();

	Navier_Stokes_Solver(int domainWidth, 
						 int domainHeight,
						 float xLength,
						 float yLength,
						 float reynoldsNumber,
						 int inkSuperSampling = 1);

	Navier_Stokes_Solver();

	void load_from_file(char *confFile);

	void reduce(char *program_name, float result[4]);
	void compute_uv_boundary();
	void compute_temperature();
	void compute_fg();
	void solve_pressure();
	//void solve_pressure_step(Render_Buffer& p_from, Render_Buffer& p_to);
	void solve_pressure_step(GLenum from, GLenum to);
	float get_residual(Render_Buffer& p, Surface_Type surface);
	void compute_new_uv();
	void adjust_time_step();
	void get_max_velocity(float &u, float &v);
	void toggle_ibfv() {
		ibfv_ = !ibfv_;
	}
	void toggle_advection() {
		advection_ = !advection_;
	}

	void get_current_max_velocity(float &u, float &v) {
		u = max_u_;
		v = max_v_;
	}

	void advect_ink();
	void clear_ink();
	void ink_grid();
	void ibfv();

	void set_obstacle_grid(const char *grid);
	void set_obstacle();

	void show_stats();
	void reset_stats();

	void set_grid_scale(float scale) {
		gridScale_ = (inkSuperSampling_ * Width()) / scale;
	}

	void set_display_field(int buffer) {
		displayBuffer_ = buffer;
	}

	void add_boundary_condition(Direction dir, Type type);
	void add_boundary_condition(Direction dir, Type type, int start, int finish);
	void add_boundary_condition(Direction dir, float u, float v, float t);
	void add_boundary_condition(Direction dir, int start, int finish, float u, float v, float t);

	void add_splat(const Splat& s) {
		splats_.push_back(s);
	};

	size_t splat_count() { return splats_.size(); }

	Splat *get_splat(float sx, float sy) {
		for (size_t i=0; i<splats_.size(); i++) {
			Splat& s = splats_[i];
			float dy = fabs(sy - s.center_[1]),
				  dx = fabs(sx - s.center_[0]),
				  d = sqrt (dy * dy + dx * dx);
			if (d < s.radius_)
				return &splats_[i];
		}
		return NULL;
	}

	void delete_splat(float sx, float sy) {
		if (splats_.size() == 0)
			return;
		for (size_t i=(size_t)splats_.size() - 1; i>=0; i--) {
			Splat& s = splats_[i];
			float dy = fabs(sy - s.center_[1]),
				  dx = fabs(sx - s.center_[0]),
				  d = sqrt (dy * dy + dx * dx);
			if (d < s.radius_) {
				splats_.erase(splats_.begin() + i);
				return;
			}
		}
	}

	void display_buffer(int buffer);

	void display();

	void splat_ink();

	void step();

	void dump();

	float Width()  { return (float)domainWidth_; }
	float Height() { return (float)domainHeight_; }
	float ReynoldsNumber() { return reynoldsNumber_; }
	float currentTime() { return totalTime_; }
	float timeStep() { return dt_; }
	bool  isDumping() { return dump_; }
};

#endif
