#include <GL/glew.h>

#include "pbuffer.h"
#include "glut_manager.h"
#include "cg_manager.h"

#include "navier_stokes.h"
#include "s_navier_stokes.h"
#include "stopwatch.hpp"

#include <iostream>
#include <fstream>
#include <vector>

using namespace std;

/***************************************************************************/

Navier_Stokes_Solver *solver;
Software_Navier_Stokes_Solver *s_solver;
Stopwatch timer;
static bool advection = true;
static bool running = true;
int fluidWindow, guiWindow;

/***************************************************************************/
// convertToDomainCoords converts coordinates of a click in window space to 
// coordinates of domain space
void convertToDomainCoords(float x, float y, float &dx, float &dy)
{
	float ww = glutGet(GLUT_WINDOW_WIDTH),
		  wh = glutGet(GLUT_WINDOW_HEIGHT);

	dx = (x/ww) * solver->Width();
	dy = ((wh-y)/wh) * solver->Height();
}

/***************************************************************************/
void display()
{
	// Clear Screen
	glClear(GL_COLOR_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);

	solver->display();

	static int frameCount = 0;

	static float old_t = 0.0f;

	frameCount++;
	if (frameCount % 20 == 0) {
		char s[100];
		float t = solver->currentTime();
		float T = timer.GetTime();


		sprintf(s, "PGHFlow @%f FPS, t=%f, t/T=%f", 20.0f / T, t, (t - old_t) / T);
		old_t = t;


		glutSetWindowTitle(s);
		timer.Reset();
	}
	
	glFlush();
	glutSwapBuffers();
}

//#define SOFTWARE
//#define HARDWARE

void idle(void)
{
	solver->step();
	
	glutPostWindowRedisplay(fluidWindow);
	glutPostRedisplay();
}

enum MENU_CHOICES {
	PAUSE_SIMULATION,
	PAUSE_ADVECTION,
	TIME_STATS,
	RESET_TIME_STATS,
	INK_CLEAR,
	INK_GRID,
	DUMP_VELOCITY,
	DUMP_PRESSURE,
	DUMP_INK,
	DISPLAY_VELOCITY_X,
	DISPLAY_VELOCITY_Y,
	DISPLAY_PRESSURE,
	DISPLAY_INK,
	DISPLAY_OBSTACLE,
	COARSE_GRID,
	MEDIUM_GRID,
	FINE_GRID,
	STEP_SIMULATION,
	TOGGLE_IBFV
};

void menu(int value)
{
	switch(value) {
	case TIME_STATS:
		solver->show_stats();
		break;
	case RESET_TIME_STATS:
		solver->reset_stats();
		break;
	case PAUSE_ADVECTION:
		solver->toggle_advection();
		break;
	case PAUSE_SIMULATION:
		running = !running;
		if (running)
			glutIdleFunc(idle);
		else
			glutIdleFunc(NULL);
		break;
	case INK_CLEAR:
		solver->clear_ink();
		break;
	case INK_GRID:
		solver->ink_grid();
		break;
	case COARSE_GRID:
		solver->set_grid_scale(4.0f);
		break;
	case MEDIUM_GRID:
		solver->set_grid_scale(8.0f);
		break;
	case FINE_GRID:
		solver->set_grid_scale(16.0f);
		break;
	case DISPLAY_VELOCITY_X:
		solver->set_display_field(VELOCITY_FIELD_X);
		break;
	case DISPLAY_VELOCITY_Y:
		solver->set_display_field(VELOCITY_FIELD_Y);
		break;
	case DISPLAY_INK:
		solver->set_display_field(INK_FIELD);
		break;
	case DISPLAY_PRESSURE:
		solver->set_display_field(PRESSURE_FIELD);
		break;
	case DISPLAY_OBSTACLE:
		solver->set_display_field(FLUID_DOMAIN);
		break;
	case DUMP_VELOCITY:
		solver->display_buffer(VELOCITY_FIELD);
		break;
	case DUMP_PRESSURE:
		solver->display_buffer(PRESSURE_FIELD);
		break;
	case DUMP_INK:
		solver->display_buffer(INK_FIELD);
		break;
	case STEP_SIMULATION:
		idle();
		break;
	case TOGGLE_IBFV:
		solver->toggle_ibfv();
		break;
	}
	glutPostRedisplay();
}

float splatColor[4];
float splatRadius;

void splat_menu(int choice)
{
	float colors[SPLAT_COLORS][4] = {
		{1.0f, 0.0f, 0.0f, 1.0f},
		{0.0f, 1.0f, 0.0f, 1.0f},
		{0.0f, 0.0f, 1.0f, 1.0f},
		{0.0f, 1.0f, 1.0f, 1.0f},
		{1.0f, 0.0f, 1.0f, 1.0f},
		{1.0f, 1.0f, 0.0f, 1.0f},
		{0.0f, 0.0f, 0.0f, 1.0f},
		{1.0f, 1.0f, 1.0f, 1.0f},
	};

	switch (choice) {
		case SPLAT_SMALL:
			splatRadius = 1.0f;
			break;
		case SPLAT_MEDIUM:
			splatRadius = 3.0f;
			break;
		case SPLAT_LARGE:
			splatRadius = 5.0f;
			break;
		case SPLAT_HUGE:
			splatRadius = solver->Width() / 5.0f;
			break;
		default:
			assert(choice < SPLAT_COLORS);
			splatColor[0] = colors[choice][0];
			splatColor[1] = colors[choice][1];
			splatColor[2] = colors[choice][2];
			splatColor[3] = colors[choice][3];
			break;
	}
}

Splat createCustomSplat(float x, float y)
{
	Splat s;
	convertToDomainCoords(x, y, s.center_[0], s.center_[1]);
	s.color_[0] = splatColor[0];
	s.color_[1] = splatColor[1];
	s.color_[2] = splatColor[2];
	s.color_[3] = splatColor[3];
	s.radius_ = splatRadius;
	return s;
}

void displayHelp()
{
	cout << "PGHFlow Help\n\nMouse:\n";

	cout << "  LMB: Places a paint splat\n";
	cout << "  Shift + LMB: drags a paint splat\n";
	cout << "  Ctrl + LMB: removes a paint splat\n";
	cout << "  MMB: paints on the ink field (acts as a transient splat)\n";
	cout << "  RMB: menu\n\nKeyboard:\n";
	cout << "  Escape: exit\n";
	cout << "  P: stop ink advection (simulation goes on)\n";
	cout << "  Space: stop simulation\n";
	cout << "  S: moves simulation forward one step\n";
	cout << "  G: draws a B&W grid on the ink\n";
	cout << "  I: toggle GPU-based IBFV\n";
	cout << "  C: clears ink field\n";
	cout << "  T: displays timing info\n";
	cout << "  R: resets timing info\n";

}

void special(int key, int x, int y)
{
	if (key == GLUT_KEY_F1) {
		displayHelp();
		return;
	}
}

void key(unsigned char k, int x, int y)
{
	if (k == 27) exit(0);
	k = toupper(k);

	if (k == 'P')
		menu(PAUSE_ADVECTION);
	if (k == 32)
		menu(PAUSE_SIMULATION);
	if (k == 'G')
		menu(INK_GRID);
	if (k == 'C')
		menu(INK_CLEAR);
	if (k == 'S')
		menu(STEP_SIMULATION);
	if (k == 'T')
		menu(TIME_STATS);
	if (k == 'R')
		menu(RESET_TIME_STATS);
	if (k == 'I')
		menu(TOGGLE_IBFV);
}

/*****************************************************************************/
// movingSplat signals that some splat is being dragged around by the user
// splatInMotion is said splat

static bool movingSplat = false;

Splat *splatInMotion;
void moveSplat(int x, int y)
{
	float dx, dy;
	convertToDomainCoords(x, y, dx, dy);
	splatInMotion->center_[0] = dx;
	splatInMotion->center_[1] = dy;
}

void mouse(int button, int state, int x, int y)
{
	int mods = glutGetModifiers();
	// start moving splat
	if (button == GLUT_LEFT_BUTTON &&
		state == GLUT_DOWN &&
		mods == GLUT_ACTIVE_SHIFT) {
		float dx, dy;
		convertToDomainCoords(x, y, dx, dy);
		Splat *s = solver->get_splat(dx, dy);
		if (s) {
			movingSplat = true;
			glutMotionFunc(moveSplat);
			splatInMotion = s;
		}
		return;
	}
	// stop moving splat
	if (button == GLUT_LEFT_BUTTON &&
		state == GLUT_UP &&
		movingSplat) {
		movingSplat = false;
		glutMotionFunc(NULL);
		splatInMotion = NULL;
		return;
	}
	// add splat
	if (button == GLUT_LEFT_BUTTON &&
		state == GLUT_UP &&
		!movingSplat &&
		mods == 0) {
			solver->add_splat(createCustomSplat(x, y));
		return;
	}
	// remove splat
	if (button == GLUT_LEFT_BUTTON &&
		state == GLUT_DOWN &&
		!movingSplat &&
		mods == GLUT_ACTIVE_CTRL) {
			float dx, dy;
			convertToDomainCoords(x, y, dx, dy);
			solver->delete_splat(dx, dy);
		return;
	}
	// transient splat
	if (button == GLUT_MIDDLE_BUTTON) {
		if (state == GLUT_DOWN) {
			float dx, dy;
			convertToDomainCoords(x, y, dx, dy);
			solver->add_splat(createCustomSplat(x, y));
			Splat *s = solver->get_splat(dx, dy);
			if (s) {
				movingSplat = true;
				glutMotionFunc(moveSplat);
				splatInMotion = s;
			}
		}
		else if (state == GLUT_UP) {
			float dx, dy;
			movingSplat = false;
			glutMotionFunc(NULL);
			splatInMotion = NULL;

			convertToDomainCoords(x, y, dx, dy);
			solver->delete_splat(dx, dy);
		}
		return;
	}
}

/***************************************************************************/

void glutSetMenus()
{
	int debug_menu = glutCreateMenu(menu);
	glutSetMenu(debug_menu);

	glutAddMenuEntry("Dump velocity field on console", DUMP_VELOCITY);
	glutAddMenuEntry("Dump pressure field on console", DUMP_PRESSURE);
	glutAddMenuEntry("Dump ink field on console", DUMP_INK);

	int grid_menu = glutCreateMenu(menu);
	glutSetMenu(grid_menu);

	glutAddMenuEntry("Coarse grid pattern", COARSE_GRID);
	glutAddMenuEntry("Medium grid pattern", MEDIUM_GRID);
	glutAddMenuEntry("Fine grid pattern", FINE_GRID);
	glutAddMenuEntry("-", 0);
	glutAddMenuEntry("[P] Toggle advection step", PAUSE_ADVECTION);
	glutAddMenuEntry("[I] Toggle IBFV", TOGGLE_IBFV);

	int display_menu = glutCreateMenu(menu);
	glutSetMenu(display_menu);

	glutAddMenuEntry("Display horizontal velocity field", DISPLAY_VELOCITY_X);
	glutAddMenuEntry("Display vertical velocity field", DISPLAY_VELOCITY_Y);
	glutAddMenuEntry("Display pressure field", DISPLAY_PRESSURE);
	glutAddMenuEntry("Display ink field", DISPLAY_INK);	
	glutAddMenuEntry("Display domain and obstacle", DISPLAY_OBSTACLE);	

	int splat_color = glutCreateMenu(splat_menu);
	glutSetMenu(splat_color);

	glutAddMenuEntry("Red",     SPLAT_RED);
	glutAddMenuEntry("Green",   SPLAT_GREEN);
	glutAddMenuEntry("Blue",    SPLAT_BLUE);
	glutAddMenuEntry("Cyan",    SPLAT_CYAN);
	glutAddMenuEntry("Magenta", SPLAT_MAGENTA);
	glutAddMenuEntry("Yellow",  SPLAT_YELLOW);
	glutAddMenuEntry("Black",   SPLAT_BLACK);
	glutAddMenuEntry("White",   SPLAT_WHITE);

	int splat_size = glutCreateMenu(splat_menu);
	glutSetMenu(splat_size);

	glutAddMenuEntry("Small",   SPLAT_SMALL);
	glutAddMenuEntry("Medium",  SPLAT_MEDIUM);
	glutAddMenuEntry("Large",   SPLAT_LARGE);
	glutAddMenuEntry("Huge",    SPLAT_HUGE);

	int m = glutCreateMenu(menu);
	glutSetMenu(m);
	glutAttachMenu(GLUT_RIGHT_BUTTON);
	
	glutAddMenuEntry("[space] Start/Stop simulation", PAUSE_SIMULATION);
	glutAddMenuEntry("[G] Write grid on ink", INK_GRID);
	glutAddMenuEntry("[C] Clear ink", INK_CLEAR);
	glutAddMenuEntry("[T] Show timing stats", TIME_STATS);
	glutAddMenuEntry("[F1] Display help on console", INK_CLEAR);
	glutAddSubMenu("Debugging", debug_menu);
	glutAddSubMenu("Ink Advection", grid_menu);
	glutAddSubMenu("Field Display", display_menu);
	glutAddSubMenu("Splat color", splat_color);
	glutAddSubMenu("Splat size",  splat_size);
}

/*****************************************************************************/

int main(int argc, char **argv)
{
	Glut_Application &app = Glut_Application::get_instance();

	Glut_Callbacks cb;
	cb.setDisplayCallback(display);
	cb.setKeyboardCallback(key);
	fluidWindow = app.addWindow(0, 0, 512, 512, "PGHFlow", cb);

	// ugly, Glut_Application is not ready yet.
	glutSetMenus();
	glutMouseFunc(mouse);
	glutSpecialFunc(special);

	GLenum err = glewInit();
	if (GLEW_OK != err) {
		cout << "Error initializing GLEW: " << glewGetErrorString(err) << endl;
		exit(1);
	} else {
		cout << "GLEW init OK" << endl;
	}

#define __CHECK_EXTENSION(a,b) \
	if (!a) { \
		cout << "Fatal: " << b << " not supported." << endl; \
		exit(1); \
	} else { \
		cout << b << " extension detected." << endl; \
	};

	__CHECK_EXTENSION(GLEW_ARB_multitexture, "GL_ARB_multitexture");
	__CHECK_EXTENSION(WGLEW_ARB_pbuffer, "WGL_ARB_pbuffer");
	__CHECK_EXTENSION(WGLEW_ARB_pixel_format, "WGL_ARB_pixel_format");
	__CHECK_EXTENSION(WGLEW_ARB_render_texture, "WGL_ARB_render_texture");
	__CHECK_EXTENSION(GLEW_NV_fence, "GL_NV_fence");
	__CHECK_EXTENSION(GLEW_NV_float_buffer, "GL_NV_float_buffer");


#undef __CHECK_EXTENSION

	glClearColor(0.3f, 0.6f, 0.9f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	string problemFile;
	solver = new Navier_Stokes_Solver();

	if (argc > 1) {
		solver->load_from_file(argv[1]);
	} else
		solver->load_from_file("pghflow.flo");

	glutSetWindow(fluidWindow);
	if (solver->Width() > solver->Height())
		glutReshapeWindow(512, 512 / ((float) solver->Width() / (float) solver->Height()));
	else
		glutReshapeWindow(512 * ((float) solver->Width() / (float) solver->Height()), 512);

	splat_menu(SPLAT_RED);
	splat_menu(SPLAT_MEDIUM);

	solver->set_display_field(INK_FIELD);

	Software_Navier_Stokes_Solver ss(3, 3, 1000.0);
	s_solver = &ss;

	timer.Start();

	if (running) 
		glutIdleFunc(idle);

	menu(PAUSE_SIMULATION);

	app.run(); // calls glutInit(), will never end
}