#include "render_buffer.h"
#include "cg_manager.h"

#include <iostream>
#include <cassert>
#include <iomanip>

using namespace std;

GLenum toGlenum(Surface_Type s)
{
	switch (s) {
		case SURFACE_FRONT: return GL_FRONT;
		case SURFACE_BACK:  return GL_BACK;
		default:
			return 0x0000;
	}
}

Surface_Type fromGLenum(GLenum s)
{
	switch (s) {
		case GL_FRONT: return SURFACE_FRONT;
		case GL_BACK:  return SURFACE_BACK;
		default:
			return SURFACE_FRONT;
	}
}

/***************************************************************************/

void Render_Buffer::draw_as_many_components(Surface_Type surface)
{
	check_errors("before Render_Buffer::draw_as_many_components()");
	int c = GetNumComponents();
	float s = 1.0f/(float)c;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glScalef(s, s, s);

	if (c == 1)
		draw_on_frame_buffer(-1,surface);
	else
		for (int i=0; i<c; i++) {
			draw_on_frame_buffer(i,surface);
			glTranslatef(1.0, 0.0, 0.0);
		}

	glPopMatrix();
}

void Render_Buffer::draw_on_frame_buffer(int component, Surface_Type surface, bool lerp)
{
	check_errors("before Render_Buffer::draw_on_frame_buffer(int component)");
	if (component != -1) {
		CG_Program pass_thru = CG_Manager::getProgram("pass-thru-normalized");
		pass_thru.enable();
		Bind(surface);
		pass_thru.setTexture("tex", *this);
		pass_thru.setParameter("modulation", 
						   component==0?1.0f:0.0f,
						   component==1?1.0f:0.0f,
						   component==2?1.0f:0.0f);
		glColor3f(1.0, 1.0, 1.0);
		glBegin(GL_QUADS);
		glTexCoord2i(0, 0);
		glVertex2i(0, 0);
		glTexCoord2i(GetWidth(), 0);
		glVertex2i(1, 0);
		glTexCoord2i(GetWidth(), GetHeight()); 
		glVertex2i(1, 1);
		glTexCoord2i(0, GetHeight());
		glVertex2i(0, 1);
		glEnd();

		Release();
		pass_thru.unsetTexture("tex");
		pass_thru.disable();
	} else {
		CG_Program pass_thru;
		if (!lerp)
			pass_thru = CG_Manager::getProgram("pass-thru");
		else
			pass_thru = CG_Manager::getProgram("lerp-pass-thru");
		pass_thru.enable();
		pass_thru.setTexture("tex", *this);
		Bind();
		glBegin(GL_QUADS);
		glTexCoord2i(0, 0);
		glVertex2i(0, 0);
		glTexCoord2i(GetWidth(), 0);
		glVertex2i(1, 0);
		glTexCoord2i(GetWidth(), GetHeight()); 
		glVertex2i(1, 1);
		glTexCoord2i(0, GetHeight());
		glVertex2i(0, 1);
		glEnd();

		Release();
		pass_thru.unsetTexture("tex");
		pass_thru.disable();

	}

	check_errors("after Render_Buffer::draw_on_frame_buffer(int component)");
}

#define MAX_DISPLAY_BUFFER_SIZE 512
static GLfloat area[MAX_DISPLAY_BUFFER_SIZE * MAX_DISPLAY_BUFFER_SIZE * 4];
void Render_Buffer::display_buffer(int comps, GLenum surface)
{
	dump_to_stream(cout, comps, surface);
}

void Render_Buffer::dump_to_stream(ostream &os, int comps, GLenum surface)
{
	check_errors("before Render_Buffer::display_buffer()");
	int w = GetWidth(),
		h = GetHeight(),
		c = GetNumComponents();

	/*Activate();
	if (surface == SURFACE_FRONT)
		glReadBuffer(GL_FRONT);
	else
		glReadBuffer(GL_BACK);
	Deactivate();*/

	switch (GetBitsPerComponent()) {
		case 8: 
		{
			//GLubyte area[100 * 100 * 4];
			assert(w < MAX_DISPLAY_BUFFER_SIZE && h < MAX_DISPLAY_BUFFER_SIZE && c <= 4);

			CopyToBuffer(area, surface);
			for (int k=0; k<comps; k++)	{
				for (int i=h-1; i>=0; i--) {
					for (int j=0; j<w; j++) {
							os/* << fixed << setprecision(4) << setw(7) << right */
								<< area[i * w * c + j * k + i] << " ";
					}
					os << endl;
				}
				os << endl;
			}
		}
		break;
		case 32:
		{
			assert(w < MAX_DISPLAY_BUFFER_SIZE && h < MAX_DISPLAY_BUFFER_SIZE && c <= 4);

			check_errors("before CopyToBuffer @ Render_Buffer::display_buffer()");
			CopyToBuffer(area, surface);
			check_errors("after CopyToBuffer @ Render_Buffer::display_buffer()");
			for (int k=0; k<comps; k++) {
				for (int i=h-1; i>=0; i--) {
					for (int j=0; j<w; j++) {
						os /*<< fixed << setprecision(4) << setw(7) << right */
							<< area[i * w * c + j * c + k] << " ";
					}
					os << endl;
				}
				os << endl;
			}
		}
		break;
		default:
			exit(1);
	};
	check_errors("after Render_Buffer::display_buffer()");
}