#include "s_navier_stokes.h"
#include <cmath>
#include <cassert>

#include <algorithm>
#include <iostream>
#include <iomanip>
using namespace std;

void Software_Navier_Stokes_Solver::inflow_west(int from, int to, int x, float fx, float fy)
{
	int f = min(from, to), t = max(from, to);
	for (; f < t; f++) {
		U[f][x] = fx;
		V[f][x] = 2*fy - V[f][x+1];
	}
}

void Software_Navier_Stokes_Solver::inflow_east(int from, int to, int x, float fx, float fy)
{
}

void Software_Navier_Stokes_Solver::inflow_north(int from, int to, int y, float fx, float fy)
{
	int f = min(from, to), t = max(from, to);
	for (; f < t; f++) {
		U[y][f] = 2 * fx - U[y-1][f];
		V[y][f] = fy;
	}
}

void Software_Navier_Stokes_Solver::inflow_south(int from, int to, int y, float fx, float fy)
{
}

void Software_Navier_Stokes_Solver::noslip_east(int from, int to, int x)
{
	int f = min(from, to), t = max(from, to);
	for (; f < t; f++) {
		U[f][x-1] = 0.0;
		V[f][x]   = -V[f][x-1];
	}
}

void Software_Navier_Stokes_Solver::noslip_west(int from, int to, int x)
{
	int f = min(from, to), t = max(from, to);
	for (; f < t; f++) {
		U[f][x] = 0.0;
		V[f][x] = -V[f][x+1];
	}
}

void Software_Navier_Stokes_Solver::noslip_north(int from, int to, int y)
{
	int f = min(from, to), t = max(from, to);
	for (; f < t; f++) {
		V[y-1][f] = 0.0;
		U[y][f]   = -U[y-1][f];
	}
}

void Software_Navier_Stokes_Solver::noslip_south(int from, int to, int y)
{
	int f = min(from, to), t = max(from, to);
	for (; f < t; f++) {
		V[y][f] = 0.0;
		U[y][f] = -U[y+1][f];
	}
}

void Software_Navier_Stokes_Solver::show_buffer(float buffer[max_buffer_size][max_buffer_size])
{
	for (int i=domainHeight-1; i>=0; i--) {
		for (int j=0; j<domainWidth; j++)
			cout /*<< fixed << setprecision(4) << setw(7) << right */
             	 << buffer[i][j] << " ";
		cout << endl;
	}
}

void Software_Navier_Stokes_Solver::compute_fg()
{
	for (int j=1; j<=jmax; j++)
		for (int i=1; i<=imax; i++) {
			float gamma = 0.9f, re = reynoldsNumber;
			float gx = 0.0f, gy = 0.0f;

			float u = U[j][i], ul = U[j][i-1], ur = U[j][i+1], uu = U[j+1][i], ud = U[j-1][i],
				  v = V[j][i], vl = V[j][i-1], vr = V[j][i+1], vu = V[j+1][i], vd = V[j-1][i],
				  uul = U[j+1][i-1], vdr = V[j-1][i+1];

			float ddudx2 = (ur-2.0f*u+ul)/(dx*dx),
				  ddvdx2 = (vr-2.0f*v+vl)/(dx*dx),
				  ddudy2 = (uu-2.0f*u+ud)/(dy*dy),
				  ddvdy2 = (vu-2.0f*v+vd)/(dy*dy);

			float mur=(u+ur)/2.0f, mul=(u+ul)/2.0f, mud=(u+ud)/2.0f, muu=(u+uu)/2.0f,
				  mvr=(v+vr)/2.0f, mvl=(v+vl)/2.0f, mvd=(v+vd)/2.0f, mvu=(v+vu)/2.0f;

			float udc1=(ul-u)/2.0f, udc2=(u-ur)/2.0f, udc3=(u-uu)/2.0f, udc4=(ud-u)/2.0f,
				  vdc1=(vl-v)/2.0f, vdc2=(v-vr)/2.0f, vdc3=(v-vu)/2.0f, vdc4=(vd-v)/2.0f,
				  dc5=(vd+vdr)/2.0f, dc6=(ul+uul)/2.0f;

			float duudx = (1.0f/dx)*(mur*mur-mul*mul)+(gamma/dx)*(fabs(mur)*udc2-fabs(mul)*udc1);
			float duvdy = (1.0f/dy)*(mvr*muu-dc5*mud)+(gamma/dy)*(fabs(mvr)*udc3-fabs(dc5)*udc4);
			float duvdx = (1.0f/dx)*(muu*mvr-dc6*mvl)+(gamma/dx)*(fabs(muu)*vdc2-fabs(dc6)*vdc1);
			float dvvdy = (1.0f/dy)*(mvu*mvu-mvd*mvd)+(gamma/dy)*(fabs(mvu)*vdc3-fabs(mvd)*vdc4);

			float f = u+dt*((1/re)*(ddudx2 + ddudy2)-duudx-duvdy+gx);
			float g = v+dt*((1/re)*(ddvdx2 + ddvdy2)-duvdx-dvvdy+gy);

			F[j][i] = f;
			G[j][i] = g;
		}

	for (int j=1; j<=jmax; j++) {
		F[j][0] = U[j][0];
		F[j][imax] = U[j][imax];
	}

	for (int i=1; i<=imax; i++) {
		G[0][i] = V[0][i];
		G[jmax][i] = V[jmax][i];
	}
}

void Software_Navier_Stokes_Solver::solve_pressure_step(SBUFFER P, SBUFFER Pp)
{
	assert(domainHeight < max_buffer_size && domainWidth < max_buffer_size);

	//for (int i=0; i<domainWidth; i++)
		//for (int j=0; j<domainHeight; j++)
			//Pp[i][j] = 0.0f;

	for (int i=1; i<=imax; i++) {
		Pp[0][i] = Pp[1][i];
		Pp[jmax+1][i] = Pp[jmax][i];
	}

	for (int j=1; j<=jmax; j++) {
		Pp[j][0] = Pp[j][1];
		Pp[j][imax+1] = Pp[j][imax];
	}

	float w = 1.0f;

	for (int i=1; i<=imax; i++)
		for (int j=1; j<=jmax; j++) {
			float eiW = (i == 1)  ? 0.0f : 1.0f;
			float eiE = (i < imax)? 1.0f : 0.0f;
			float ejS = (j == 1)  ? 0.0f : 1.0f;
			float ejN = (j < jmax)? 1.0f : 0.0f;

			float sr = (1 - w) * P[j][i];
			float factor = w / ((eiE + eiW)/(dx*dx) + 
								(ejN + ejS)/(dy*dy));
			float t1 = (eiE * P[j][i+1] + eiW * P[j][i-1])/(dx*dx);
			float t2 = (ejN * P[j+1][i] + ejS * P[j-1][i])/(dy*dy);

			float rhs = (1/dt)*((F[j][i] - F[j][i-1])/dx + (G[j][i] - G[j-1][i])/dy);

			Pp[j][i] = sr + factor * (t1 + t2 - rhs);
		}

	/*for (int i=0; i<domainWidth; i++)
		for (int j=0; j<domainHeight; j++)
			P[j][i] = Pp[j][i];*/
}

void Software_Navier_Stokes_Solver::compute_new_uv()
{
	for (int i=1; i<=imax; i++)
		for (int j=1; j<=imax; j++) {
			U[j][i] = F[j][i] - (dt/dx) * (P[j][i+1] - P[j][i]);
			V[j][i] = G[j][i] - (dt/dy) * (P[j+1][i] - P[j][i]);
		}
}

float Software_Navier_Stokes_Solver::get_residual(SBUFFER P)
{
	float r = 0.0f;
	// Compute residual
	for (int i=1; i<=imax; i++)
		for (int j=1; j<=jmax; j++) {
			float eiW = (i == 1)  ? 0.0f : 1.0f;
			float eiE = (i < imax)? 1.0f : 0.0f;
			float ejS = (j == 1)  ? 0.0f : 1.0f;
			float ejN = (j < jmax)? 1.0f : 0.0f;

			float t1 = eiE * (P[j][i+1] - P[j][i]),
				  t2 = eiW * (P[j][i]   - P[j][i-1]),
				  t3 = ejN * (P[j+1][i] - P[j][i]),
				  t4 = ejS * (P[j][i]   - P[j-1][i]);

			float rhs = (1/dt)*((F[j][i] - F[j][i-1])/dx + (G[j][i] - G[j-1][i])/dy);

			float v = (t1 - t2)/(dx*dx) + (t3 - t4)/(dy*dy) - rhs;

			R[j][i] = v * v;

			r += R[j][i];
		}

	return sqrt(r / (float)(imax*jmax));
}

void Software_Navier_Stokes_Solver::solve_pressure()
{
	static int jacobi_steps = 16;
	static int counter = 0;
	static bool turbo_mode = false;	
	
	int maxiter = 32;
	int turbo_cycle = 128;
	float eps = 1e-2;



	while (1) {
		for (int i=0; i<jacobi_steps; i++) {
			solve_pressure_step((pressure_s1 == SURFACE_FRONT)?P:P2, (pressure_s2 == SURFACE_FRONT)?P:P2);
			swap(pressure_s1, pressure_s2);
		}

		if (turbo_mode) {
			counter++;
			if (counter == turbo_cycle) {
				counter = 0;
				float f = get_residual((pressure_s1 == SURFACE_FRONT)?P:P2);
				if (f > eps) {
					turbo_mode = false;
				}
			}
		} else {
			float f = get_residual((pressure_s1 == SURFACE_FRONT)?P:P2);
			if (f > eps) {
				jacobi_steps *= 2;
				counter = 0;
				if (jacobi_steps > maxiter) {
					jacobi_steps = maxiter;
#ifdef _DEBUG
					cerr << "Punting out of solve_pressure with sub-epsilon r: " << f << endl;
#endif
					break;
				}
			} else {
				jacobi_steps /= 4;
				if (jacobi_steps < 4) {
					jacobi_steps = 4;
					counter++;
					if (counter == 128) {
						// Achieved steady state?
						cout << "Achieved steady state in " << timer.GetTime() << endl;
#ifdef _DEBUG
						cout << "Entering turbo mode!" << endl;
#endif
						turbo_mode = true;
						counter = 0;
					}
				}
#ifdef _DEBUG
				cout << "Ok, steps = " << jacobi_steps << ", r: " << f << endl;
#endif
				break;
			}
		}
	}
}

void Software_Navier_Stokes_Solver::adjust_time_step()
{
	float u = fabs(U[1][1]), v = fabs(V[1][1]);

	for (int i=1; i<=imax; i++)
		for (int j=1; j<=jmax; j++) {
			u = max(u, fabs(U[i][j]));
			v = max(v, fabs(V[i][j]));
		}

	dt = min(min(dx/u, dy/v), (reynoldsNumber/2.0f) / (1.0f/(dx*dx) + 1.0f/(dy*dy)));
}

void Software_Navier_Stokes_Solver::display_buffer(int b)
{
	switch (b) {
		case 0:
			cout << "uv" << endl;
			show_buffer(U);
			cout << endl;
			show_buffer(V);
			break;
		case 1:
			break;
		case 2:
			cout << "fg" << endl;
			show_buffer(F);
			cout << endl;
			show_buffer(G);
			break;
		case 3:
			cout << "pressure" << endl;
			show_buffer(P);
			break;
		case 4:
			cout << "residual" << endl;
			show_buffer(R);
			break;
	}
}

void Software_Navier_Stokes_Solver::set_uv_boundary()
{
	noslip_north(1, imax, jmax+1);
	noslip_south(1, imax, 0);
	noslip_west (1, jmax, 0);
	noslip_east (1, jmax, imax+1);

	inflow_north(1, imax, jmax+1, 1.0f, 0.0f);
}
